from __future__ import annotations

import networkx as nx

from swmmanywhere.metric_utilities import metrics


@metrics.register
def new_metric(synthetic_G: nx.Graph, real_G: nx.Graph, **kwargs) -> float:
    """New metric function."""
    return len(synthetic_G.edges) / len(real_G.edges)



"""The main SWMManywhere module to generate and run a synthetic network."""

from __future__ import annotations

import importlib
from pathlib import Path

import geopandas as gpd
import jsonschema
import pandas as pd
import pyswmm
from tqdm.auto import tqdm

import swmmanywhere.geospatial_utilities as go
from swmmanywhere import filepaths, parameters, preprocessing
from swmmanywhere.graph_utilities import (
    iterate_graphfcns,
    load_graph,
    save_graph,
    validate_graphfcn_list,
)
from swmmanywhere.logging import logger, verbose
from swmmanywhere.metric_utilities import iterate_metrics, validate_metric_list
from swmmanywhere.post_processing import synthetic_write
from swmmanywhere.utilities import yaml_dump, yaml_load


def _check_defaults(config: dict) -> dict:
    """Check the config for needed values and add them from defaults if missing.

    Args:
        config (dict): The configuration.

    Returns:
        dict: The configuration with defaults added.
    """
    config_ = load_config(validation=False)
    for key in ["run_settings", "graphfcn_list", "metric_list"]:
        if key not in config:
            config[key] = config_[key]

    return config


def swmmanywhere(config: dict) -> tuple[Path, dict | None]:
    """Run SWMManywhere processes.

    This function runs the SWMManywhere processes, including downloading data,
    preprocessing the graphfcns, running the model, and comparing the results
    to real data using metrics. The function will always return the path to
    the generated .inp file. If real data (either a results file or the .inp,
    as well as graph, and subcatchments) is provided, the function will also
    return the metrics comparing the synthetic network with the real.

    Args:
        config (dict): The loaded config as a dict.

    Returns:
        tuple[Path, dict | None]: The address of generated .inp and metrics.
    """
    # Check for defaults
    config = _check_defaults(config)

    # Currently precipitation must be provided via address_overrides, otherwise
    # the default storm.dat file will be used
    if not Path(
        config.get("address_overrides", {}).get(
            "precipitation", Path("precipitation.dat")
        )
    ).exists():
        config["address_overrides"] = config.get("address_overrides", {})
        config["address_overrides"]["precipitation"] = (
            Path(__file__).parent / "defs" / "storm.dat"
        )

    # Create the project structure
    logger.info("Creating project structure.")
    addresses = filepaths.FilePaths(
        config["base_dir"],
        config["project"],
        config["bbox"],
        config.get("bbox_number", None),
        config.get("model_number", None),
        config.get("extension", "parquet"),
        **config.get("address_overrides", {}),
    )

    logger.info(f"Project structure created at {addresses.project_paths.base_dir}")
    logger.info(f"Project name: {config['project']}")
    logger.info(
        f"""Bounding box: {config["bbox"]}, 
                number: {addresses.bbox_paths.bbox_number}"""
    )
    logger.info(f"Model number: {addresses.model_paths.model_number}")

    # Save config file
    if verbose():
        save_config(config, addresses.model_paths.model / "config.yml")

    # Load the parameters and perform any manual overrides
    logger.info("Loading and setting parameters.")
    params = parameters.get_full_parameters()
    for category, overrides in config.get("parameter_overrides", {}).items():
        for key, val in overrides.items():
            logger.info(f"Setting {category} {key} to {val}")
            setattr(params[category], key, val)

    # If `allowable_networks` has been changed, force a redownload of street graph.
    if "allowable_networks" in config.get("parameter_overrides", {}).get(
        "topology_derivation", {}
    ):
        logger.info("Allowable networks have been changed, removing old street graph.")
        addresses.bbox_paths.street.unlink(missing_ok=True)

    # Run downloads
    logger.info("Running downloads.")
    preprocessing.run_downloads(
        config["bbox"],
        addresses,
        network_types=params["topology_derivation"].allowable_networks,
    )

    # Identify the starting graph
    logger.info("Iterating graphs.")
    if config.get("starting_graph", None):
        G = load_graph(config["starting_graph"])
    else:
        G = preprocessing.create_starting_graph(addresses)

    # Validate the graphfcn order
    validate_graphfcn_list(config["graphfcn_list"], G)

    # Iterate the graph functions
    logger.info("Iterating graph functions.")
    G = iterate_graphfcns(G, config["graphfcn_list"], params, addresses)

    # Save the final graph
    logger.info("Saving final graph and writing inp file.")
    go.graph_to_geojson(
        G, addresses.model_paths.nodes, addresses.model_paths.edges, G.graph["crs"]
    )
    save_graph(G, addresses.model_paths.graph)

    # Check any edges
    if len(G.edges) == 0:
        logger.warning("No edges in graph, returning graph file.")
        return addresses.model_paths.graph, None

    # Write to .inp
    synthetic_write(addresses)

    # Return model if no run
    if not config.get("run_model", True):
        return addresses.model_paths.inp, None

    # Run the model
    logger.info("Running the synthetic model.")
    synthetic_results = run(addresses.model_paths.inp, **config["run_settings"])
    logger.info("Writing synthetic results.")
    if verbose():
        synthetic_results.to_parquet(addresses.model_paths.model / "results.parquet")

    # Get the real results
    if config.get("real", {}).get("results", None):
        logger.info("Loading real results.")
        real_results = pd.read_parquet(config["real"]["results"])
    elif config.get("real", {}).get("inp", None):
        logger.info("Running the real model.")
        real_results = run(config["real"]["inp"], **config["run_settings"])
        if verbose():
            real_results.to_parquet(
                config["real"]["inp"].parent / "real_results.parquet"
            )
    else:
        logger.info("No real network provided, returning SWMM .inp file.")
        return addresses.model_paths.inp, None

    # Iterate the metrics
    logger.info("Iterating metrics.")
    if addresses.model_paths.subcatchments.suffix == ".geoparquet":
        subs = gpd.read_parquet(addresses.model_paths.subcatchments)
    else:
        subs = gpd.read_file(addresses.model_paths.subcatchments)

    if config["real"]["subcatchments"].suffix == ".geoparquet":
        real_subs = gpd.read_parquet(config["real"]["subcatchments"])
    else:
        real_subs = gpd.read_file(config["real"]["subcatchments"])
    metrics = iterate_metrics(
        synthetic_results,
        subs,
        G,
        real_results,
        real_subs,
        load_graph(config["real"]["graph"]),
        config["metric_list"],
        params["metric_evaluation"],
    )
    logger.info("Metrics complete")
    return addresses.model_paths.inp, metrics


def check_top_level_paths(config: dict):
    """Check the top level paths (`base_dir`) in the config.

    Args:
        config (dict): The configuration.

    Raises:
        FileNotFoundError: If a top level path does not exist.
    """
    key = "base_dir"
    if not Path(config[key]).exists():
        raise FileNotFoundError(f"{key} not found at {config[key]}")
    config[key] = Path(config[key])
    return config


def check_address_overrides(config: dict):
    """Check the address overrides in the config.

    Args:
        config (dict): The configuration.

    Raises:
        FileNotFoundError: If an address override path does not exist.
    """
    overrides = config.get("address_overrides", None)

    if not overrides:
        return config

    for key, path in overrides.items():
        if not Path(path).exists():
            raise FileNotFoundError(f"{key} not found at {path}")
        config["address_overrides"][key] = Path(path)
    return config


def check_real_network_paths(config: dict):
    """Check the paths to the real network in the config.

    Args:
        config (dict): The configuration.

    Raises:
        FileNotFoundError: If a real network path does not exist.
    """
    real = config.get("real", None)

    if not real:
        return config

    for key, path in real.items():
        if not isinstance(path, str):
            continue
        if not Path(path).exists():
            raise FileNotFoundError(f"{key} not found at {path}")
        config["real"][key] = Path(path)

    return config


def check_starting_graph(config: dict):
    """Check the starting graph in the config.

    Args:
        config (dict): The configuration.

    Raises:
        FileNotFoundError: If the starting graph path does not exist.
    """
    # If no starting graph, return
    if not config.get("starting_graph", None):
        return config

    # Check the starting graph exists and convert to Path
    config["starting_graph"] = Path(config["starting_graph"])
    if not config["starting_graph"].exists():
        raise FileNotFoundError(
            f"""starting_graph not found at 
                                {config["starting_graph"]}"""
        )

    return config


def check_parameter_overrides(config: dict):
    """Check the parameter overrides in the config.

    Args:
        config (dict): The configuration.

    Raises:
        ValueError: If a parameter override is not in the parameters
            dictionary.
    """
    params = parameters.get_full_parameters()
    for category, overrides in config.get("parameter_overrides", {}).items():
        if category not in params:
            raise ValueError(
                f"""{category} not a category of parameter. Must
                             be one of {params.keys()}."""
            )

        # Get the available properties for a category
        cat_properties = params[category].model_json_schema()["properties"]

        for key, val in overrides.items():
            # Check that the parameter is available
            if key not in cat_properties:
                raise ValueError(f"{key} not found in {category}.")

    return config


def import_module(module: Path):
    """Import module with importlib.

    Args:
        module (Path): path to module.
    """
    # Import the module
    spec = importlib.util.spec_from_file_location(  # type: ignore[attr-defined]
        module.stem, module
    )
    module = importlib.util.module_from_spec(spec)  # type: ignore[attr-defined]
    spec.loader.exec_module(module)


def import_modules(modules: list[str | Path]):
    """Import modules specified in list of files.

    Args:
        modules (list): List of files

    Raises:
        ValueError: If a module does not exist.
    """
    for module in modules:
        module = Path(module)

        # Check that module exists
        if not module.exists():
            raise FileNotFoundError(f"Module not found at {module}")
        import_module(module)


def check_and_register_custom_graphfcns(config: dict):
    """Check, register and validate custom graphfcns in the config.

    Args:
        config (dict): The configuration.

    Raises:
        ValueError: If a graphfcn module does not exist.
        ValueError: If a custom graphfcn is not successfully registered.
    """
    import_modules(config.get("custom_graphfcn_modules", []))

    # Validate the import
    validate_graphfcn_list(config.get("graphfcn_list", []))

    return config


def check_and_register_custom_metrics(config: dict):
    """Check, register and validate custom metrics in the config.

    Args:
        config (dict): The configuration.

    Raises:
        ValueError: If the custom metrics module does not exist.
    """
    import_modules(config.get("custom_metric_modules", []))

    # Validate metric list
    validate_metric_list(config.get("metric_list", []))

    return config


def register_custom_parameters(config: dict):
    """Register custom parameter modules.

    Args:
        config (dict): The configuration.
    """
    import_modules(config.get("custom_parameter_modules", []))
    return config


def save_config(config: dict, config_path: Path):
    """Save the configuration to a file.

    Args:
        config (dict): The configuration.
        config_path (Path): The path to save the configuration.
    """
    yaml_dump(config, config_path.open("w"))


def load_config(
    config_path: Path = Path(__file__).parent / "defs" / "demo_config.yml",
    validation: bool = True,
    schema_fid: Path | None = None,
):
    """Load, validate, and convert Paths in a configuration file.

    Note, if using a custom graphfcn, load_config must be called with validation=True.

    Args:
        config_path (Path): The path to the configuration file.
        validation (bool, optional): Whether to validate the configuration.
            Defaults to True.
        schema_fid (Path, optional): The path to the schema file. Defaults to
            None.

    Returns:
        dict: The configuration.
    """
    # Load the schema
    schema_fid = (
        Path(__file__).parent / "defs" / "schema.yml"
        if schema_fid is None
        else Path(schema_fid)
    )
    schema = yaml_load(schema_fid.read_text())

    # Load the config
    config = yaml_load(config_path.read_text())

    if not validation:
        return config

    # Validate the config
    jsonschema.validate(instance=config, schema=schema)

    # Check top level paths
    config = check_top_level_paths(config)

    # Check address overrides
    config = check_address_overrides(config)

    # Check real network paths
    config = check_real_network_paths(config)

    # Check starting graph
    config = check_starting_graph(config)

    # Check parameter overrides
    config = check_parameter_overrides(config)

    # Check and register custom metrics
    config = check_and_register_custom_metrics(config)

    # Check and register custom graphfcns
    config = check_and_register_custom_graphfcns(config)

    # Register custom parameters
    config = register_custom_parameters(config)

    return config


def run(
    model: Path,
    reporting_iters: int = 50,
    duration: int = 86400,
    storevars: list[str] = ["flooding", "flow"],
):
    """Run a SWMM model and store the results.

    Args:
        model (Path): The path to the SWMM model .inp file.
        reporting_iters (int, optional): The number of iterations between
            storing results. Defaults to 50.
        duration (int, optional): The duration of the simulation in seconds.
            Starts at the 'START_DATE' and 'START_TIME' defined in the 'model'
            .inp file Defaults to 86400.
        storevars (list[str], optional): The variables to store. Defaults to
            ['flooding','flow'].

    Returns:
        pd.DataFrame: A DataFrame containing the results.
    """
    with pyswmm.Simulation(str(model)) as sim:
        sim.start()
        logger.info(f"{model} initialised in pyswmm")

        # Define the variables to store
        variables = {
            "flooding": {"class": pyswmm.Nodes, "id": "_nodeid"},
            "depth": {"class": pyswmm.Nodes, "id": "_nodeid"},
            "flow": {"class": pyswmm.Links, "id": "_linkid"},
            "runoff": {"class": pyswmm.Subcatchments, "id": "_subcatchmentid"},
        }

        results_list = []
        for var, info in variables.items():
            if var not in storevars:
                continue
            # Rather than calling eg Nodes or Links, only call them if they
            # are needed for storevars because they carry a significant
            # overhead
            pobjs = info["class"](sim)
            results_list += [
                {"object": x, "variable": var, "id": info["id"]} for x in pobjs
            ]

        # Iterate the model
        results = []
        t_ = sim.current_time
        ind = 0
        logger.info(f"Starting simulation for: {model}")

        progress_bar = tqdm(total=duration, disable=not verbose())

        offset = 0
        while (
            (offset <= duration)
            & (sim.current_time < sim.end_time)
            & (not sim._terminate_request)
        ):
            progress_bar.update((sim.current_time - t_).total_seconds() - offset)
            offset = (sim.current_time - t_).total_seconds()

            ind += 1

            # Iterate the main model timestep
            time = sim._model.swmm_step()

            # Break condition
            if time < 0:
                sim._terminate_request = True
                break

            # Check whether to save results
            if ind % reporting_iters != 1:
                continue

            # Store results in a list of dictionaries
            for storevar in results_list:
                results.append(
                    {
                        "date": sim.current_time,
                        "value": getattr(storevar["object"], storevar["variable"]),
                        "variable": storevar["variable"],
                        "id": getattr(storevar["object"], storevar["id"]),
                    }
                )

    logger.info("Model run complete.")
    return pd.DataFrame(results)



"""Utility functions for shortest path algorithms."""

from __future__ import annotations

import heapq
from collections import defaultdict
from typing import Hashable

import networkx as nx


def tarjans_pq(
    G: nx.MultiDiGraph, root: int | str, weight_attr: str = "weight"
) -> nx.MultiDiGraph:
    """Tarjan's algorithm for a directed minimum spanning tree.

    Also known as a minimum spanning arborescence, this algorithm finds the
    minimum directed spanning tree rooted at a given vertex (root) in a directed
    graph.

    Args:
        G (nx.MultiDiGraph): The input graph.
        root (int | str): The root node (i.e., that all vertices in the graph
            should flow to).
        weight_attr (str): The name of the edge attribute containing the edge
            weights. Defaults to 'weight'.

    Returns:
        nx.MultiDiGraph: The directed minimum spanning tree.
    """
    # Copy the graph and relabel the nodes
    G_ = G.copy()
    new_nodes = {node: i for i, node in enumerate(G.nodes)}
    node_mapping = {i: node for i, node in enumerate(G.nodes)}
    G_ = nx.relabel_nodes(G_, new_nodes)

    # Extract the new root label, edges and weights
    root = new_nodes[root]
    edges = [(u, v, d[weight_attr]) for u, v, d in G_.edges(data=True)]

    # Initialize data structures
    graph = defaultdict(list)
    for u, v, weight in edges:
        graph[v].append((u, weight))

    n = len(G.nodes)
    parent = {}  # Parent pointers for the MST
    in_edge_pq: list = []  # Priority queue to store incoming edges

    # Initialize the priority queue with edges incoming to the root
    for u, weight in graph[root]:
        heapq.heappush(in_edge_pq, (weight, u, root))

    mst_edges = []
    mst_weight = 0
    outfalls: dict = {}
    while in_edge_pq:
        weight, u, v = heapq.heappop(in_edge_pq)

        if u not in parent:
            # If v is not in the MST yet, add the edge (u, v) to the MST
            parent[u] = v
            mst_edges.append((u, v))
            mst_weight += weight

            if v in outfalls:
                outfalls[u] = outfalls[v]

            elif G_.get_edge_data(u, v)[0]["edge_type"] == "outfall":
                outfalls[u] = node_mapping[u]

            # Add incoming edges to v to the priority queue
            for w, weight_new in graph[u]:
                heapq.heappush(in_edge_pq, (weight_new, w, u))

    # Check if all vertices are reachable from the root
    if len(parent) != n - 1:
        raise ValueError("Graph is not connected or has multiple roots.")

    new_graph = nx.MultiDiGraph()

    for u, v in mst_edges:
        d = G_.get_edge_data(u, v)[0]
        new_graph.add_edge(u, v, **d)

    for u, d in G_.nodes(data=True):
        new_graph.nodes[u].update(d)

    nx.set_node_attributes(new_graph, outfalls, "outfall")
    new_graph = nx.relabel_nodes(new_graph, node_mapping)
    new_graph.graph = G.graph.copy()
    return new_graph


def dijkstra_pq(
    G: nx.MultiDiGraph, outfalls: list, weight_attr: str = "weight"
) -> nx.MultiDiGraph:
    """Dijkstra's algorithm for shortest paths to outfalls.

    This function calculates the shortest paths from each node in the graph to
    the nearest outfall. The graph is modified to include the outfall
    and the shortest path length.

    Args:
        G (nx.MultiDiGraph): The input graph.
        outfalls (list): A list of outfall nodes.
        weight_attr (str): The name of the edge attribute containing the edge
            weights. Defaults to 'weight'.

    Returns:
        nx.MultiDiGraph: The graph with the shortest paths to outfalls.
    """
    G = G.copy()
    # Initialize the dictionary with infinity for all nodes
    shortest_paths = {node: float("inf") for node in G.nodes}

    # Initialize the dictionary to store the paths
    paths: dict[Hashable, list] = {node: [] for node in G.nodes}

    # Set the shortest path length to 0 for outfalls
    for outfall in outfalls:
        shortest_paths[outfall] = 0
        paths[outfall] = [outfall]

    # Initialize a min-heap with (distance, node) tuples
    heap = [(0, outfall) for outfall in outfalls]
    while heap:
        # Pop the node with the smallest distance
        dist, node = heapq.heappop(heap)

        # For each neighbor of the current node
        for neighbor, _, edge_data in G.in_edges(node, data=True):
            # Calculate the distance through the current node
            alt_dist = dist + edge_data[weight_attr]
            # If the alternative distance is shorter

            if alt_dist >= shortest_paths[neighbor]:
                continue

            # Update the shortest path length
            shortest_paths[neighbor] = alt_dist
            # Update the path
            paths[neighbor] = paths[node] + [neighbor]
            # Push the neighbor to the heap
            heapq.heappush(heap, (alt_dist, neighbor))

    # Remove nodes with no path to an outfall
    for node in [node for node, path in paths.items() if not path]:
        G.remove_node(node)
        del paths[node], shortest_paths[node]

    if len(G.nodes) == 0:
        raise ValueError("""No nodes with path to outfall, """)

    edges_to_keep: set = set()

    for path in paths.values():
        # Assign outfall
        outfall = path[0]
        for node in path:
            G.nodes[node]["outfall"] = outfall
            G.nodes[node]["shortest_path"] = shortest_paths[node]

        # Store path
        edges_to_keep.update(zip(path[1:], path[:-1]))

    # Remove edges not on paths
    new_graph = G.copy()
    for u, v in G.edges():
        if (u, v) not in edges_to_keep:
            new_graph.remove_edge(u, v)

    return new_graph



"""**********
Exceptions
**********

Custom exceptions for NetComp.
"""

from __future__ import annotations


class UndefinedException(Exception):
    """Raised when matrix to be returned is undefined"""


class InputError(Exception):
    """Raised when input to algorithm is invalid"""


class KathleenError(Exception):
    """Raised when food is gross or it is too cold"""



"""**********
Resistance
**********

Resistance matrix. Renormalized version, as well as conductance and commute matrices.
"""

from __future__ import annotations

from contextlib import suppress

import networkx as nx
import numpy as np
from numpy import linalg as la
from scipy import linalg as spla
from scipy.sparse import issparse

from ..exception import UndefinedException
from .matrices import laplacian_matrix


def resistance_matrix(A, check_connected=True):
    """Return the resistance matrix of G.

    Parameters
    ----------
    A : NumPy matrix or SciPy sparse matrix
        Adjacency matrix of a graph.

    check_connected : Boolean, optional (default=True)
        If false, then the resistance matrix will be computed even for
        disconnected matrices. See Notes.

    Returns:
    -------
    R : NumPy matrix
       Matrix of pairwise resistances between nodes.

    Notes:
    -----
    Uses formula for resistance matrix R in terms of Moore-Penrose of
    pseudoinverse (non-normalized) graph Laplacian. See e.g. Theorem 2.1 in [1].

    This formula can be computed even for disconnected graphs, although the
    interpretation in this case is unclear. Thus, the usage of
    check_connected=False is recommended only to reduce computation time in a
    scenario in which the user is confident the graph in question is, in fact,
    connected.

    Since we do not expect the pseudoinverse of the laplacian to be sparse, we
    convert L to dense form before running np.linalg.pinv(). The returned
    resistance matrix is dense.

    See Also:
    --------
    nx.laplacian_matrix

    References:
    ----------
    .. [1] W. Ellens, et al. (2011)
       Effective graph resistance.
       Linear Algebra and its Applications, 435 (2011)

    """
    n, m = A.shape
    # check if graph is connected
    if check_connected:
        if issparse(A):
            G = nx.from_scipy_sparse_array(A)
        else:
            G = nx.from_numpy_array(A)
        if not nx.is_connected(G):
            raise UndefinedException(
                "Graph is not connected. Resistance matrix is undefined."
            )
    L = laplacian_matrix(A)
    with suppress(Exception):
        L = L.todense()

    M = la.pinv(L)
    # calculate R in terms of M
    d = np.reshape(np.diag(M), (n, 1))
    ones = np.ones((n, 1))
    R = np.dot(d, ones.T) + np.dot(ones, d.T) - M - M.T
    return R


def commute_matrix(A):
    """Return the commute matrix of the graph associated with adj. matrix A.

    Parameters
    ----------
    A : NumPy matrix or SciPy sparse matrix
        Adjacency matrix of a graph.

    Returns:
    -------
    C : NumPy matrix
       Matrix of pairwise resistances between nodes.

    Notes:
    -----
    Uses formula for commute time matrix in terms of resistance matrix,

    C = R*2*|E|

    where |E| is the number of edges in G. See e.g. Theorem 2.8 in [1].

    See Also:
    --------
    laplacian_matrix
    resistance_matrix

    References:
    ----------
    .. [1] W. Ellens, et al. (2011)
       Effective graph resistance.
       Linear Algebra and its Applications, 435 (2011)

    """
    R = resistance_matrix(A)
    E = A.sum() / 2  # number of edges in graph
    C = 2 * E * R
    return C


def renormalized_res_mat(A, beta=1):
    """Return the renormalized resistance matrix of graph associated with A.

    To renormalize a resistance R, we apply the function

    R' = R / (R + beta)

    In this way, the renormalized resistance of nodes in disconnected components
    is 1. The parameter beta determines the penalty for disconnection. If we set
    beta to be approximately the maximum resistance found in the network, then
    the penalty for disconnection is at least 1/2.

    Parameters
    ----------
    A : NumPy matrix or SciPy sparse matrix
        Adjacency matrix of a graph.

    nodelist : list, optional
       The rows and columns are ordered according to the nodes in nodelist. If
       nodelist is None, then the ordering is produced by G.nodes().

    weight : string or None, optional (default='weight')
       The edge data key used to compute each value in the matrix.
       If None, then each edge has weight 1.

    beta : float, optional
       Scaling parameter in renormalization. Must be greater than or equal to
       1. Determines how heavily disconnection is penalized.

    Returns:
    -------
    R :  NumPy array
       Matrix of pairwise renormalized resistances between nodes.

    Notes:
    -----
    This function converts to a NetworkX graph, as it uses the algorithms
    therein for identifying connected components.

    See Also:
    --------
    resistance_matrix

    """
    if issparse(A):
        G = nx.from_scipy_sparse_array(A)
    else:
        G = nx.from_numpy_array(A)

    if isinstance(G, nx.Graph):
        cc = nx.connected_components
    else:
        cc = nx.weakly_connected_components

    n = len(G)
    subgraphR = []
    for c in cc(G):
        subgraph = G.subgraph(c).copy()
        a_sub = nx.adjacency_matrix(subgraph)
        r_sub = resistance_matrix(a_sub)
        subgraphR.append(r_sub)
    R = spla.block_diag(*subgraphR)
    # now, resort R so that it matches the original node list
    component_order = []
    for component in cc(G):
        component_order += list(component)
    component_order = list(np.argsort(component_order))
    R = R[component_order, :]
    R = R[:, component_order]
    renorm = np.vectorize(lambda r: r / (r + beta))
    R = renorm(R)
    # set resistance for different components to 1
    R[R == 0] = 1
    R = R - np.eye(n)  # don't want diagonal to be 1
    return R


def conductance_matrix(A):
    """Return the conductance matrix of G.

    The conductance matrix of G is the element-wise inverse of the resistance
    matrix. The diagonal is set to 0, although it is formally infinite. Nodes in
    disconnected components have 0 conductance.

    Parameters
    ----------
    G : graph
       A NetworkX graph

    nodelist : list, optional
       The rows and columns are ordered according to the nodes in nodelist. If
       nodelist is None, then the ordering is produced by G.nodes().

    weight : string or None, optional (default='weight')
       The edge data key used to compute each value in the matrix.
       If None, then each edge has weight 1.

    Returns:
    -------
    C :  NumPy array
       Matrix of pairwise conductances between nodes.


    See Also:
    --------
    resistance_matrix
    renormalized_res_mat

    """
    if issparse(A):
        G = nx.from_scipy_sparse_array(A)
    else:
        G = nx.from_numpy_array(A)
    if isinstance(G, nx.Graph):
        cc = nx.connected_components
    else:
        cc = nx.weakly_connected_components
    subgraphC = []
    for c in cc(G):
        subgraph = G.subgraph(c).copy()
        a_sub = nx.adjacency_matrix(subgraph)
        r_sub = resistance_matrix(a_sub)
        m = len(subgraph)
        # add one to diagonal, invert, remove one from diagonal:
        c_sub = 1 / (r_sub + np.eye(m)) - np.eye(m)
        subgraphC.append(c_sub)
    C = spla.block_diag(*subgraphC)
    # resort C so that it matches the original node list
    component_order = []
    for component in cc(G):
        component_order += list(component)
    component_order = list(np.argsort(component_order))
    C = C[component_order, :]
    C = C[:, component_order]
    return C



"""**********
Eigenstuff
**********

Functions for calculating eigenstuff of graphs.
"""

from __future__ import annotations

from contextlib import suppress

import numpy as np
from numpy import linalg as la
from scipy import sparse as sps
from scipy.sparse import issparse
from scipy.sparse import linalg as spla

from .matrices import _eps, _flat

######################
## Helper Functions ##
######################


def _eigs(M, which="SR", k=None):
    """Helper function for getting eigenstuff.

    Parameters
    ----------
    M : matrix, numpy or scipy sparse
        The matrix for which we hope to get eigenstuff.
    which : string in {'SR','LR'}
        If 'SR', get eigenvalues with smallest real part. If 'LR', get largest.
    k : int
        Number of eigenvalues to return

    Returns:
    -------
    evals, evecs : numpy arrays
        Eigenvalues and eigenvectors of matrix M, sorted in ascending or
        descending order, depending on 'which'.

    See Also:
    --------
    numpy.linalg.eig
    scipy.sparse.eigs
    """
    n, _ = M.shape
    if k is None:
        k = n
    if which not in ["LR", "SR"]:
        raise ValueError("which must be either 'LR' or 'SR'.")
    M = M.astype(float)
    if issparse(M) and k < n - 1:
        evals, evecs = spla.eigs(M, k=k, which=which)
    else:
        with suppress(Exception):
            M = M.todense()

        evals, evecs = la.eig(M)
        # sort dem eigenvalues
        inds = np.argsort(evals)
        if which == "LR":
            inds = inds[::-1]
        else:
            pass
        inds = inds[:k]
        evals = evals[inds]
        evecs = np.matrix(evecs[:, inds])
    return np.real(evals), np.real(evecs)


#####################
##  Get Eigenstuff ##
#####################


def normalized_laplacian_eig(A, k=None):
    """Return the eigenstuff of the normalized Laplacian matrix of graph
    associated with adjacency matrix A.

    Calculates via eigenvalues if

    K = D^(-1/2) A D^(-1/2)

    where `A` is the adjacency matrix and `D` is the diagonal matrix of
    node degrees. Since L = I - K, the eigenvalues and vectors of L can
    be easily recovered.

    Parameters
    ----------
    A : NumPy matrix
        Adjacency matrix of a graph

    k : int, 0 < k < A.shape[0]-1
        The number of eigenvalues to grab.

    Returns:
    -------
    lap_evals : NumPy array
       Eigenvalues of L

    evecs : NumPy matrix
       Columns are the eigenvectors of L

    Notes:
    -----
    This way of calculating the eigenvalues of the normalized graph laplacian is
    more numerically stable than simply forming the matrix L = I - K and doing
    numpy.linalg.eig on the result. This is because the eigenvalues of L are
    close to zero, whereas the eigenvalues of K are close to 1.

    References:
    ----------

    See Also:
    --------
    nx.laplacian_matrix
    nx.normalized_laplacian_matrix
    """
    n, m = A.shape
    ##
    ## TODO: implement checks on the adjacency matrix
    ##
    degs = _flat(A.sum(axis=1))
    # the below will break if
    inv_root_degs = [d ** (-1 / 2) if d > _eps else 0 for d in degs]
    inv_rootD = sps.spdiags(inv_root_degs, [0], n, n, format="csr")
    # build normalized diffusion matrix
    K = inv_rootD * A * inv_rootD
    evals, evecs = _eigs(K, k=k, which="LR")
    lap_evals = 1 - evals
    return np.real(lap_evals), np.real(evecs)



"""Tests for the main module."""

from __future__ import annotations

import os
import tempfile
from pathlib import Path

import geopandas as gpd
import jsonschema
import pytest
import yaml

from swmmanywhere import parameters, swmmanywhere
from swmmanywhere.defs import copy_test_data
from swmmanywhere.graph_utilities import graphfcns
from swmmanywhere.metric_utilities import metrics
from swmmanywhere.utilities import plot_basic, plot_map


def test_run():
    """Test the run function."""
    demo_dir = Path(__file__).parent.parent / "src" / "swmmanywhere" / "defs"
    model = demo_dir / "basic_drainage_all_bits.inp"
    storevars = ["flooding", "flow", "runoff", "depth"]
    results = swmmanywhere.run(model, reporting_iters=50, storevars=storevars)
    assert set(results.variable.unique()) == set(storevars)

    # Ensure more reporting iterations results in more results
    results_ = swmmanywhere.run(model, reporting_iters=25, storevars=storevars)
    assert results_.shape[0] > results.shape[0]

    # Ensure a shorter duration results in fewer results
    results_ = swmmanywhere.run(model, duration=10000, storevars=storevars)
    assert results_.shape[0] < results.shape[0]

    model.with_suffix(".out").unlink()
    model.with_suffix(".rpt").unlink()


@pytest.mark.parametrize("run", [True, False])
def test_swmmanywhere(run, wbt_zip_path):
    """Test the swmmanywhere function."""
    with tempfile.TemporaryDirectory(dir=".") as temp_dir:
        # Load the config
        test_data_dir = Path(__file__).parent / "test_data"
        defs_dir = Path(__file__).parent.parent / "src" / "swmmanywhere" / "defs"
        with (defs_dir / "demo_config.yml").open("r") as f:
            config = yaml.safe_load(f)

        # Set some test values
        base_dir = Path(temp_dir)
        config["base_dir"] = str(base_dir)
        config["bbox"] = [0.05677, 51.55656, 0.07193, 51.56726]
        config["address_overrides"] = {
            "building": str(test_data_dir / "building.geoparquet"),
            "whiteboxtools_binaries_zip": str(wbt_zip_path),
        }
        config["parameter_overrides"] = {
            "subcatchment_derivation": {"subbasin_streamorder": 5}
        }
        config["run_settings"]["duration"] = 1000
        config["model_number"] = 0

        # Fill the real dict with unused paths to avoid filevalidation errors
        config["real"]["subcatchments"] = str(defs_dir / "storm.dat")
        config["real"]["inp"] = str(defs_dir / "storm.dat")
        config["real"]["graph"] = str(defs_dir / "storm.dat")

        # Write the config
        with open(base_dir / "test_config.yml", "w") as f:
            yaml.dump(config, f)

        # Load and test validation of the config
        config = swmmanywhere.load_config(base_dir / "test_config.yml")

        # Set the test config to just use the generated data
        model_dir = base_dir / "demo" / "bbox_1" / "model_0"
        config["real"]["subcatchments"] = model_dir / "subcatchments.geoparquet"
        config["real"]["inp"] = model_dir / "model_0.inp"
        config["real"]["graph"] = model_dir / "graph.parquet"
        config["run_model"] = run

        # Run swmmanywhere
        os.environ["SWMMANYWHERE_VERBOSE"] = "true"
        inp, metrics = swmmanywhere.swmmanywhere(config)

        # Check what was generated
        assert (inp.parent / f"{config['graphfcn_list'][-1]}_graph.json").exists()
        assert inp.exists()

        assert (inp.parent / "edges.geoparquet").exists()
        assert (inp.parent / "nodes.geoparquet").exists()

        # Check cost separately since it isn't implicitly checked by running the model
        edges = gpd.read_file(inp.parent / "edges.geoparquet")
        assert "cost_usd" in edges.columns
        assert edges.cost_usd.min() > 0

        if not run:
            assert inp is not None
            assert inp.exists()
            assert metrics is None
            return None

        # Check metrics were calculated
        assert metrics is not None
        for key, val in metrics.items():
            if not val:
                continue
            assert isinstance(val, float)

        assert set(metrics.keys()) == set(config["metric_list"])

        # Check results were saved
        assert (inp.parent / "results.parquet").exists()
        assert (config["real"]["inp"].parent / "real_results.parquet").exists()

        # Check the map functions
        plot_basic(inp.parent / "nodes.geoparquet", inp.parent / "edges.geoparquet")
        plot_map(inp.parent)


def test_load_config_file_validation():
    """Test the file validation of the config."""
    with tempfile.TemporaryDirectory(dir=".") as temp_dir:
        defs_dir = Path(__file__).parent.parent / "src" / "swmmanywhere" / "defs"
        base_dir = Path(temp_dir)

        # Test file not found
        with pytest.raises(FileNotFoundError) as exc_info:
            swmmanywhere.load_config(base_dir / "test_config.yml")
            assert "test_config.yml" in str(exc_info.value)

        with (defs_dir / "demo_config.yml").open("r") as f:
            config = yaml.safe_load(f)

        # Correct and avoid filevalidation errors
        config["real"] = None

        # Fill with unused paths to avoid filevalidation errors
        config["base_dir"] = str(defs_dir / "storm.dat")

        with open(base_dir / "test_config.yml", "w") as f:
            yaml.dump(config, f)

        config = swmmanywhere.load_config(base_dir / "test_config.yml")
        assert isinstance(config, dict)


def test_load_config_schema_validation():
    """Test the schema validation of the config."""
    with tempfile.TemporaryDirectory(dir=".") as temp_dir:
        defs_dir = Path(__file__).parent.parent / "src" / "swmmanywhere" / "defs"
        base_dir = Path(temp_dir)

        # Load the config
        with (defs_dir / "demo_config.yml").open("r") as f:
            config = yaml.safe_load(f)

        # Make an edit not to schema
        config["base_dir"] = 1

        with open(base_dir / "test_config.yml", "w") as f:
            yaml.dump(config, f)

        # Test schema validation
        with pytest.raises(jsonschema.exceptions.ValidationError) as exc_info:
            swmmanywhere.load_config(base_dir / "test_config.yml")
            assert "null" in str(exc_info.value)


def test_save_config():
    """Test the save_config function."""
    with tempfile.TemporaryDirectory(dir=".") as temp_dir:
        temp_dir = Path(temp_dir)
        defs_dir = Path(__file__).parent.parent / "src" / "swmmanywhere" / "defs"

        with (defs_dir / "demo_config.yml").open("r") as f:
            config = yaml.safe_load(f)

        # Correct and avoid filevalidation errors
        config["real"] = None

        # Fill with unused paths to avoid filevalidation errors
        config["base_dir"] = str(defs_dir / "storm.dat")

        swmmanywhere.save_config(config, temp_dir / "test.yml")

        # Reload to check OK
        config = swmmanywhere.load_config(temp_dir / "test.yml")


@pytest.mark.downloads
def test_minimal_req():
    """Test SWMManywhere with minimal info."""
    with tempfile.TemporaryDirectory(dir=".") as temp_dir:
        config = {
            "base_dir": Path(temp_dir),
            "project": "my_test",
            "bbox": [1.52740, 42.50524, 1.54273, 42.51259],
        }

        swmmanywhere.swmmanywhere(config)


def test_custom_metric():
    """Test adding a custom metric."""
    with tempfile.TemporaryDirectory(dir=".") as temp_dir:
        # Load the config
        gf_module = str(Path(__file__).parent / "test_data" / "custom_metrics.py")

        config = swmmanywhere.load_config(validation=False)
        config["custom_metric_modules"] = [str(gf_module)]
        config["metric_list"].append("new_metric")

        # Set some test values
        config_address = Path(temp_dir) / "test_config.yml"
        config["base_dir"] = temp_dir
        config["bbox"] = [0, 1, 0, 1]
        del config["real"]

        # Write the config
        swmmanywhere.save_config(config, config_address)

        # Import of `custom_metric` by CI testing environment adds `new_metric`
        if "new_metric" in metrics:
            del metrics["new_metric"]

        # Load and test validation of the config
        config = swmmanywhere.load_config(config_address)

        # Check metric was added
        assert "new_metric" in metrics

        # Remove the custom metric for other tests
        del metrics["new_metric"]


def test_custom_graphfcn():
    """Test adding a custom graphfcn."""
    with tempfile.TemporaryDirectory(dir=".") as temp_dir:
        # Load the config
        gf_module = str(Path(__file__).parent / "test_data" / "custom_graphfcns.py")

        config = swmmanywhere.load_config(validation=False)
        config["custom_graphfcn_modules"] = [str(gf_module)]
        config["graphfcn_list"].append("new_graphfcn")

        # Set some test values
        config_address = Path(temp_dir) / "test_config.yml"
        config["base_dir"] = temp_dir
        config["bbox"] = [0, 1, 0, 1]
        del config["real"]

        # Write the config
        swmmanywhere.save_config(config, config_address)

        # Import of `custom_graphfcns` by CI testing environment adds `new_graphfcn`
        if "new_graphfcn" in graphfcns:
            del graphfcns["new_graphfcn"]

        # Load and test validation of the config
        config = swmmanywhere.load_config(config_address)

        # Check graphfcn was added
        assert "new_graphfcn" in graphfcns

        # Remove the custom graphfcn for other tests
        del graphfcns["new_graphfcn"]


def test_custom_parameters(tmp_path):
    """Test custom parameters provided by filename."""
    config = swmmanywhere.load_config(validation=False)
    config["custom_graphfcn_modules"] = [
        Path(__file__).parent / "test_data" / "custom_parameters.py"
    ]

    config_address = tmp_path / "test_config.yml"
    config["base_dir"] = tmp_path
    config["bbox"] = [0, 1, 0, 1]
    del config["real"]

    # Write the config
    swmmanywhere.save_config(config, config_address)

    # Load and test validation of the config
    config = swmmanywhere.load_config(config_address)

    # Check graphfcn was added
    assert "new_params" in parameters.get_full_parameters()


def test_copy_test_data(tmp_path):
    """Test the copy_test_data function."""
    # Load test data
    copy_test_data(tmp_path)

    assert (tmp_path / "storm.dat").exists()
    assert (tmp_path / "bellinge_small.inp").exists()
    assert (tmp_path / "bellinge_small_graph.json").exists()
    assert (tmp_path / "bellinge_small_subcatchments.geojson").exists()
    assert (tmp_path / "nodes.geojson").exists()
    assert (tmp_path / "edges.geojson").exists()
    swmmanywhere.run(str(tmp_path / "bellinge_small.inp"))



"""Module for graphfcns that identify outfalls."""

from __future__ import annotations

from typing import Dict

import networkx as nx
import pandas as pd
import shapely

from swmmanywhere import geospatial_utilities as go
from swmmanywhere import parameters
from swmmanywhere.graph_utilities import BaseGraphFunction, register_graphfcn
from swmmanywhere.logging import logger


def _get_points(
    G: nx.Graph,
) -> tuple[Dict[str, shapely.Point], Dict[str, shapely.Point]]:
    """Get the river and street points from the graph.

    A river point are the start and end nodes of an edge with `edge_type = river`.
    A street point are the start and end nodes of an edge with
    `edge_type = street`.

    Args:
        G (nx.Graph): A graph

    Returns:
        river_points (dict): A dictionary of river points
        street_points (dict): A dictionary of street points
    """
    # Get edge types, convert to nx.Graph to remove keys
    etypes = nx.get_edge_attributes(nx.Graph(G), "edge_type")

    # Get river and street points as a dict
    n_types = (
        pd.DataFrame(etypes.items(), columns=["key", "type"])
        .explode("key")
        .reset_index(drop=True)
        .groupby("type")["key"]
        .apply(list)
        .to_dict()
    )
    river_points = {
        n: shapely.Point(G.nodes[n]["x"], G.nodes[n]["y"])
        for n in n_types.get("river", {})
    }
    street_points = {
        n: shapely.Point(G.nodes[n]["x"], G.nodes[n]["y"]) for n in n_types["street"]
    }

    return river_points, street_points


def _pair_rivers(
    G: nx.Graph,
    river_points: Dict[str, shapely.Point],
    street_points: Dict[str, shapely.Point],
    river_buffer_distance: float,
    outfall_length: float,
) -> nx.Graph:
    """Pair river and street nodes.

    Pair river and street nodes within a certain distance of each other. If
    there are no plausible outfalls for an entire subgraph, then a dummy river
    node is created and the lowest elevation node is used as the outfall.

    Args:
        G (nx.Graph): A graph
        river_points (dict): A dictionary of river points
        street_points (dict): A dictionary of street points
        river_buffer_distance (float): The distance within which a river and
            street node can be paired
        outfall_length (float): The length of the outfall

    Returns:
        G (nx.Graph): A graph
    """
    matched_outfalls = {}

    # Insert a dummy river node and use lowest elevation node as outfall
    # for each subgraph with no matched outfalls
    subgraphs = []
    for sg in nx.weakly_connected_components(G):
        sg = G.subgraph(sg).copy()
        subgraphs.append(sg)

        # Ignore the rivers, they are drained later
        if all([d["edge_type"] == "river" for _, _, d in sg.edges(data=True)]):
            continue

        # Pair up the river and street nodes for each subgraph
        street_points_ = {k: v for k, v in street_points.items() if k in sg.nodes}

        subgraph_outfalls = go.nearest_node_buffer(
            street_points_, river_points, river_buffer_distance
        )

        # Check if there are any matched outfalls
        if subgraph_outfalls:
            # Update all matched outfalls
            matched_outfalls.update(subgraph_outfalls)
            continue

        # In cases of e.g., an area with no rivers to discharge into or too
        # small a buffer

        # Identify the lowest elevation node
        lowest_elevation_node = min(
            sg.nodes, key=lambda x: sg.nodes[x]["surface_elevation"]
        )

        # Create a dummy river to discharge into
        name = f"{lowest_elevation_node}-dummy_river"
        dummy_river = {
            "id": name,
            "x": G.nodes[lowest_elevation_node]["x"] + 1,
            "y": G.nodes[lowest_elevation_node]["y"] + 1,
        }
        sg.add_node(name)
        nx.set_node_attributes(sg, {name: dummy_river})

        # Update function's dicts
        matched_outfalls[lowest_elevation_node] = name
        river_points[name] = shapely.Point(dummy_river["x"], dummy_river["y"])

        logger.warning(
            f"""No outfalls found for subgraph containing 
                        {lowest_elevation_node}, using this node as outfall."""
        )

    G = nx.compose_all(subgraphs)

    # Add edges between the paired river and street nodes
    for street_id, river_id in matched_outfalls.items():
        # TODO instead use a weight based on the distance between the two nodes
        G.add_edge(
            street_id,
            river_id,
            **{
                "length": outfall_length,
                "weight": outfall_length,
                "edge_type": "outfall",
                "geometry": shapely.LineString(
                    [street_points[street_id], river_points[river_id]]
                ),
                "id": f"{street_id}-{river_id}-outfall",
            },
        )

    return G


def _root_nodes(G: nx.Graph) -> nx.Graph:
    """Root nodes with a waste node.

    Connect all nodes that have nowhere to flow to to a waste node, i.e., the
    root of the entire graph.

    Args:
        G (nx.Graph): A graph

    Returns:
        G (nx.Graph): A graph
    """
    G_ = G.copy()

    # Add edges from the river nodes to a waste node
    G.add_node("waste")

    for node in G_.nodes:
        if G.out_degree(node) == 0:
            # Location of the waste node doesn't matter - so if there
            # are multiple river nodes with out_degree 0 - that's fine.
            G.nodes["waste"]["x"] = G.nodes[node]["x"] + 1
            G.nodes["waste"]["y"] = G.nodes[node]["y"] + 1
            G.add_edge(
                node,
                "waste",
                **{
                    "length": 0,
                    "weight": 0,
                    "edge_type": "waste-outfall",
                    "id": f"{node}-waste-outfall",
                },
            )
    return G


def _connect_mst_outfalls(paired_G: nx.Graph, raw_G: nx.Graph) -> nx.Graph:
    """Connect outfalls to a waste node.

    Run a minimum spanning tree (MST) on the paired graph to identify the
    'efficient' outfalls. These outfalls are inserted into the original graph.

    Args:
        paired_G (nx.Graph): A graph where streets and rivers are paired with
            outfalls.
        raw_G (nx.Graph): A graph where streets and rivers are separated

    Returns:
        (nx.Graph): A graph
    """
    # Find shortest path to identify only 'efficient' outfalls. The MST
    # makes sense here over shortest path as each node is only allowed to
    # be visited once - thus encouraging fewer outfalls. In shortest path
    # nodes near rivers will always just pick their nearest river node.
    T = nx.minimum_spanning_tree(paired_G.to_undirected(), weight="length")

    # Retain the shortest path outfalls in the original graph
    for u, v, d in T.edges(data=True):
        if (d["edge_type"] == "outfall") & (v != "waste") & (u != "waste"):
            if u not in raw_G.nodes():
                raw_G.add_node(u, **paired_G.nodes[u])
            elif v not in raw_G.nodes():
                raw_G.add_node(v, **paired_G.nodes[v])

            # Need to check both directions since T is undirected
            if (u, v) in paired_G.edges():
                raw_G.add_edge(u, v, **d)
            elif (v, u) in paired_G.edges():
                raw_G.add_edge(v, u, **d)
            else:
                raise ValueError(f"Edge {u}-{v} not found in paired_G")

    return raw_G


@register_graphfcn
class identify_outfalls(
    BaseGraphFunction,
    required_edge_attributes=["length", "edge_type"],
    required_node_attributes=["x", "y", "surface_elevation"],
):
    """identify_outfalls class."""

    def __call__(
        self, G: nx.Graph, outfall_derivation: parameters.OutfallDerivation, **kwargs
    ) -> nx.Graph:
        """Identify outfalls in a combined river-street graph.

        This function identifies outfalls in a combined river-street graph. An
        outfall is a node that is connected to a river and a street. Each street
        node is paired with the nearest river node provided that it is within
        a distance of outfall_derivation.river_buffer_distance - this provides a
        large set of plausible outfalls. If there are no plausible outfalls for an
        entire subgraph, then a dummy river node is created and the lowest
        elevation node is paired with it. Any street->river/outfall link is given
        a `weight` and `length` of outfall_derivation.outfall_length, this is to
        ensure some penalty on the total number of outfalls selected.

        Two methods are available for determining which plausible outfalls to
        retain:

        - `withtopo`: all plausible outfalls are retained, connected to a single
        'waste' node and assumed to be later identified as part of the
        `derive_topology` graphfcn.

        - `separate`: the retained outfalls are those that are selected as part
        of the minimum spanning tree (MST) of the combined street-river graph.
        This method can be temporamental because the MST is undirected, because
        rivers are inherently directed unusual outfall locations may be retained.

        Args:
            G (nx.Graph): A graph
            outfall_derivation (parameters.OutfallDerivation): An OutfallDerivation
                parameter object
            **kwargs: Additional keyword arguments are ignored.

        Returns:
            G (nx.Graph): A graph
        """
        G = G.copy()

        river_points, street_points = _get_points(G)

        G_ = _pair_rivers(
            G,
            river_points,
            street_points,
            outfall_derivation.river_buffer_distance,
            outfall_derivation.outfall_length,
        )

        # Set the length of the river edges to 0 - from a design perspective
        # once water is in the river we don't care about the length - since it
        # costs nothing
        for _, _, d in G_.edges(data=True):
            if d["edge_type"] == "river":
                d["length"] = 0
                d["weight"] = 0

        # Add edges from the river nodes to a waste node
        G_ = _root_nodes(G_)

        if outfall_derivation.method == "withtopo":
            # The outfalls can be derived as part of the shortest path calculations
            return G_
        elif outfall_derivation.method == "separate":
            return _connect_mst_outfalls(G_, G)
        else:
            raise ValueError(f"Unknown method {outfall_derivation.method}")



from __future__ import annotations

from pathlib import Path

import pytest
from pywbt.pywbt import _get_platform_suffix


def pytest_collection_modifyitems(config, items):
    """Skip tests marked with downloads."""
    if not config.getoption("markexpr", "False"):
        config.option.markexpr = "not downloads"


@pytest.fixture
def wbt_zip_path() -> Path:
    """Determine the platform specific binary for whiteboxtools for testing.

    All WBT binaries are stored in `wbt_zip` within the `tests` directory.

    Based on implementation in [`pywbt`](https://github.com/cheginit/pywbt).
    """
    _, suffix, _ = _get_platform_suffix()

    return Path(__file__).parent / "wbt_zip" / f"WhiteboxTools_{suffix}.zip"



"""Parameters module for SWMManywhere."""

from __future__ import annotations

from typing import Callable

import numpy as np
from pydantic import BaseModel, Field, model_validator

from swmmanywhere.logging import logger

parameter_register = {}


def register_parameter_group(name: str) -> Callable:
    """Register a parameter group.

    Args:
        name (str): Name of the parameter group that it will be keyed to in
            parameter_register.
    """

    def wrapper(cls: BaseModel) -> BaseModel:
        if name in parameter_register:
            logger.warning(f"{name} already in parameter register, overwriting.")
        parameter_register[name] = cls()
        return cls

    return wrapper


def get_full_parameters():
    """Get the full set of parameters."""
    return parameter_register


def get_full_parameters_flat():
    """Get the full set of parameters in a flat format."""
    parameters = get_full_parameters()
    # Flatten
    # parameters_flat = {k : {**y, **{'category' : cat}}
    #                    for cat,v in parameters.items()
    #                     for k, y in v.model_json_schema()['properties'].items()}
    parameters_flat = {}
    for cat, v in parameters.items():
        for k, y in v.model_json_schema()["properties"].items():
            parameters_flat[k] = {**y, **{"category": cat}}

    return parameters_flat


@register_parameter_group(name="subcatchment_derivation")
class SubcatchmentDerivation(BaseModel):
    """Parameters for subcatchment derivation."""

    subbasin_streamorder: int = Field(
        default=None,
        ge=1,
        le=20,
        description="Stream order for subbasin derivation.",
        json_schema_extra={"unit": "-"},
    )

    subbasin_membership: float = Field(
        default=0.5,
        ge=0,
        le=1,
        json_schema_extra={"unit": "-"},
        description="Membership threshold for subbasin derivation.",
    )

    subbasin_clip_method: str = Field(
        default="subbasin",
        json_schema_extra={"unit": "-"},
        description="Method to clip subbasins, can be `subbasin` or `community`.",
    )

    lane_width: float = Field(
        default=3.5,
        ge=2.0,
        le=5.0,
        json_schema_extra={"unit": "m"},
        description="Width of a road lane.",
    )

    carve_depth: float = Field(
        default=2.0,
        ge=1.0,
        le=3.0,
        json_schema_extra={"unit": "m"},
        description="Depth of road/river carve for flow accumulation.",
    )

    max_street_length: float = Field(
        default=60.0,
        ge=40.0,
        le=100.0,
        json_schema_extra={"unit": "m"},
        description="Distance to split streets into segments.",
    )

    node_merge_distance: float = Field(
        default=10,
        ge=1,
        le=39.9,  # should be less than max_street_length
        json_schema_extra={"unit": "m"},
        description="Distance within which to merge street nodes.",
    )


@register_parameter_group(name="outfall_derivation")
class OutfallDerivation(BaseModel):
    """Parameters for outfall derivation."""

    method: str = Field(
        default="separate",
        json_schema_extra={"unit": "-"},
        description="""Method to derive outfall locations, 
            can be 'separate' or 'withtopo'.""",
    )

    river_buffer_distance: float = Field(
        default=150.0,
        ge=10.0,
        le=500.0,
        json_schema_extra={"unit": "m"},
        description="Buffer distance to link rivers to streets.",
    )

    outfall_length: float = Field(
        default=40.0,
        ge=0.0,
        le=600.0,
        json_schema_extra={"unit": "-"},
        description="Weight to discourage street drainage into river buffers.",
    )


@register_parameter_group(name="topology_derivation")
class TopologyDerivation(BaseModel):
    """Parameters for topology derivation."""

    allowable_networks: list = Field(
        default=["walk", "drive"],
        min_length=1,
        json_schema_extra={"unit": "-"},
        description="OSM networks to consider",
    )

    weights: list = Field(
        default=["chahinian_slope", "chahinian_angle", "length", "contributing_area"],
        min_length=1,
        json_schema_extra={"unit": "-"},
        description="Weights for topo derivation",
    )

    omit_edges: list = Field(
        default=["motorway", "motorway_link", "bridge", "tunnel", "corridor"],
        min_length=1,
        json_schema_extra={"unit": "-"},
        description="OSM paths pipes are not allowed under",
    )

    chahinian_slope_scaling: float = Field(
        default=1,
        le=1,
        ge=0,
        json_schema_extra={"unit": "-"},
        description="Constant to apply to surface slope in topo derivation",
    )

    chahinian_angle_scaling: float = Field(
        default=0,
        le=1,
        ge=0,
        json_schema_extra={"unit": "-"},
        description="Constant to apply to chahinian angle in topo derivation",
    )

    length_scaling: float = Field(
        default=0.1,
        le=1,
        ge=0,
        json_schema_extra={"unit": "-"},
        description="Constant to apply to length in topo derivation",
    )

    contributing_area_scaling: float = Field(
        default=0.1,
        le=1,
        ge=0,
        json_schema_extra={"unit": "-"},
        description="Constant to apply to contributing area in topo derivation",
    )

    chahinian_slope_exponent: float = Field(
        default=1,
        le=2,
        ge=0,
        json_schema_extra={"unit": "-"},
        description="Exponent to apply to surface slope in topo derivation",
    )

    chahinian_angle_exponent: float = Field(
        default=1,
        le=2,
        ge=0,
        json_schema_extra={"unit": "-"},
        description="Exponent to apply to chahinian angle in topo derivation",
    )

    length_exponent: float = Field(
        default=1,
        le=2,
        ge=0,
        json_schema_extra={"unit": "-"},
        description="Exponent to apply to length in topo derivation",
    )

    contributing_area_exponent: float = Field(
        default=1,
        le=2,
        ge=0,
        json_schema_extra={"unit": "-"},
        description="Exponent to apply to contributing area in topo derivation",
    )

    @model_validator(mode="after")
    def check_weights(self) -> TopologyDerivation:
        """Check that weights have associated scaling and exponents."""
        for weight in self.weights:
            if not hasattr(self, f"{weight}_scaling"):
                raise ValueError(f"Missing {weight}_scaling")
            if not hasattr(self, f"{weight}_exponent"):
                raise ValueError(f"Missing {weight}_exponent")
        return self


@register_parameter_group("hydraulic_design")
class HydraulicDesign(BaseModel):
    """Parameters for hydraulic design."""

    diameters: list = Field(
        default=np.linspace(0.15, 3, int((3 - 0.15) / 0.075) + 1).tolist(),
        min_length=1,
        json_schema_extra={"unit": "m"},
        description="""Diameters to consider in 
                            pipe by pipe method""",
    )
    max_fr: float = Field(
        default=0.8,
        le=1,
        ge=0,
        json_schema_extra={"unit": "-"},
        description="Maximum filling ratio in pipe by pipe method",
    )
    min_shear: float = Field(
        default=2,
        le=3,
        ge=0,
        json_schema_extra={"unit": "Pa"},
        description="Minimum wall shear stress in pipe by pipe method",
    )
    min_v: float = Field(
        default=0.75,
        le=2,
        ge=0,
        json_schema_extra={"unit": "m/s"},
        description="Minimum velocity in pipe by pipe method",
    )
    max_v: float = Field(
        default=5,
        le=10,
        ge=3,
        json_schema_extra={"unit": "m/s"},
        description="Maximum velocity in pipe by pipe method",
    )
    min_depth: float = Field(
        default=0.5,
        le=1,
        ge=0,
        json_schema_extra={"unit": "m"},
        description="Minimum excavation depth in pipe by pipe method",
    )
    max_depth: float = Field(
        default=5,
        le=10,
        ge=2,
        json_schema_extra={"unit": "m"},
        description="Maximum excavation depth in pipe by pipe method",
    )
    precipitation: float = Field(
        default=0.006,
        le=0.010,
        ge=0.001,
        description="Depth of design storm in pipe by pipe method",
        json_schema_extra={"unit": "m"},
    )
    depth_nbins: int = Field(
        default=10,
        ge=1,
        json_schema_extra={"unit": "-"},
        description="Number of bins to discretise depth for in pipe by pipe method",
    )
    edge_design_parameters: list[str] = Field(
        default=["diameter", "cost_usd"],
        min_length=1,
        json_schema_extra={"unit": "-"},
        description="""Edge parameters calculated by the design process to retain in the
                    graph after the pipe_by_pipe graphfcn has been applied.""",
    )


@register_parameter_group(name="metric_evaluation")
class MetricEvaluation(BaseModel):
    """Parameters for metric evaluation."""

    grid_scale: float = Field(
        default=100,
        le=5000,
        ge=10,
        json_schema_extra={"unit": "m"},
        description="Scale of the grid for metric evaluation",
    )

    warmup: float = Field(
        default=0,
        ge=0,
        lt=1,
        json_schema_extra={"unit": "fraction"},
        description="""Warmup period as a fraction of the total simulation time. This
            is used to exclude the initial part of the simulation from the metric
            calculations.""",
    )



from __future__ import annotations

import networkx as nx

from swmmanywhere.graph_utilities import BaseGraphFunction, register_graphfcn


@register_graphfcn
class new_graphfcn(BaseGraphFunction, adds_edge_attributes=["new_attrib"]):
    """New graphfcn class."""

    def __call__(self, G: nx.graph, **kwargs) -> nx.Graph:
        """Adds new_attrib to the graph."""
        G = G.copy()
        nx.set_edge_attributes(G, "new_value", "new_attrib")
        return G



from __future__ import annotations

import difflib
import filecmp
import shutil
import tempfile
from pathlib import Path

import geopandas as gpd
import pandas as pd
import pyswmm
from shapely import geometry as sgeom

from swmmanywhere import post_processing as stt
from swmmanywhere.filepaths import FilePaths

fid = (
    Path(__file__).parent.parent
    / "src"
    / "swmmanywhere"
    / "defs"
    / "basic_drainage_all_bits.inp"
)


def test_overwrite_section():
    """Test the overwrite_section function.

    All this tests is that the data is written to the file.
    """
    # Copy the example file to a temporary file

    with tempfile.TemporaryDirectory(dir=".") as temp_dir:
        temp_fname = Path(temp_dir) / "temp.inp"
        shutil.copy(fid, temp_fname)
        data = [
            ["1", "1", "1", "1.166", "100", "500", "0.5", "0", "empty"],
            ["2", "1", "1", "1.1", "100", "500", "0.5", "0", "empty"],
            ["subca_3", "1", "1", "2", "100", "400", "0.5", "0", "empty"],
        ]

        section = "[SUBCATCHMENTS]"
        stt.overwrite_section(data, section, temp_fname)
        with temp_fname.open("r") as file:
            content = file.read()
        assert "subca_3" in content


def test_change_flow_routing():
    """Test the change_flow_routing function."""
    # Copy the example file to a temporary file
    with tempfile.TemporaryDirectory(dir=".") as temp_dir:
        temp_fid = Path(temp_dir) / "temp.inp"
        shutil.copy(fid, temp_fid)
        new_routing = "STEADY"
        stt.change_flow_routing(new_routing, temp_fid)
        with temp_fid.open("r") as file:
            content = file.read()
        assert "STEADY" in content


def test_data_input_dict_to_inp():
    """Test the data_input_dict_to_inp function.

    All this tests is that the data is written to a new file.
    """
    data_dict = {
        "SUBCATCHMENTS": [
            ["1", "1", "1", "1.166", "100", "500", "0.5", "0", "empty"],
            ["2", "1", "1", "1.1", "100", "500", "0.5", "0", "empty"],
            ["subca_3", "1", "1", "2", "100", "400", "0.5", "0", "empty"],
        ]
    }

    with tempfile.TemporaryDirectory(dir=".") as temp_dir:
        temp_fid = Path(temp_dir) / "temp.inp"
        shutil.copy(fid, temp_fid)
        stt.data_dict_to_inp(data_dict, fid, temp_fid)
        with temp_fid.open("r") as file:
            content = file.read()
        assert "subca_3" in content


def test_explode_polygon():
    """Test the explode_polygon function."""
    df = pd.Series(
        {
            "subcatchment": "1",
            "geometry": sgeom.Polygon([(0, 0), (1, 0), (1, 1), (0, 1)]),
        }
    )
    result = stt.explode_polygon(df)
    assert result.shape[0] == 5
    assert result.loc[3, "y"] == 1


def generate_data_dict():
    """Generate a data dict for testing."""
    nodes = gpd.GeoDataFrame(
        {
            "id": ["node1", "node2"],
            "x": [0, 1],
            "y": [0, 1],
            "max_depth": [1, 1],
            "chamber_floor_elevation": [1, 1],
            "surface_elevation": [2, 2],
            "manhole_area": [0.5, 0.5],
        }
    )
    outfalls = gpd.GeoDataFrame(
        {
            "id": ["node2_outfall"],
            "chamber_floor_elevation": [-4],
            "surface_elevation": [-4],
            "x": [-49],
            "y": [-49],
        }
    )
    conduits = gpd.GeoDataFrame(
        {
            "id": ["node1-node2", "node2-node2_outfall"],
            "u": ["node1", "node2"],
            "v": ["node2", "node2_outfall"],
            "length": [1, (50**2 + 50**2) ** 0.5],
            "roughness": [0.01, 0.01],
            "shape_swmm": ["CIRCULAR", "CIRCULAR"],
            "diameter": [1, 1],
            "capacity": [1e10, 1e10],
        }
    )
    subs = gpd.GeoDataFrame(
        {
            "subcatchment": ["node1-sub"],
            "rain_gage": ["1"],
            "id": ["node1"],
            "area": [1],
            "rc": [1],
            "width": [1],
            "slope": [0.001],
            "geometry": [sgeom.Polygon([(0, 0), (1, 0), (1, 1), (0, 1)])],
        }
    )
    rain_fid = "storm.dat"
    event = {"name": "1", "unit": "mm", "interval": "00:05", "fid": rain_fid}
    symbol = {"x": 0, "y": 0, "name": "1"}

    return {
        "nodes": nodes,
        "outfalls": outfalls,
        "conduits": conduits,
        "subs": subs,
        "event": event,
        "symbol": symbol,
    }


def test_synthetic_write():
    """Test the synthetic_write function."""
    data_dict = generate_data_dict()
    with tempfile.TemporaryDirectory(dir=".") as model_dir:
        model_dir = Path(model_dir)
        addresses = FilePaths(
            base_dir=model_dir,
            bbox_bounds=[0, 1, 0, 1],
            project_name="test",
            extension="json",
            precipitation="storm.dat",
        )
        # Write the model with synthetic_write
        nodes = gpd.GeoDataFrame(data_dict["nodes"])
        nodes.geometry = gpd.points_from_xy(nodes.x, nodes.y)
        nodes.to_file(addresses.model_paths.nodes)
        nodes = nodes.set_index("id")
        edges = gpd.GeoDataFrame(pd.DataFrame(data_dict["conduits"]).iloc[[0]])
        edges.geometry = [
            sgeom.LineString([nodes.loc[u, "geometry"], nodes.loc[v, "geometry"]])
            for u, v in zip(edges.u, edges.v)
        ]
        edges.to_file(addresses.model_paths.edges)
        subs = data_dict["subs"].copy()
        subs["subcatchment"] = ["node1"]
        subs.to_file(addresses.model_paths.subcatchments)
        stt.synthetic_write(addresses)

        # Write the model with data_dict_to_inp
        comparison_file = addresses.model_paths.model / "model_base.inp"
        template_fid = (
            Path(__file__).parent.parent
            / "src"
            / "swmmanywhere"
            / "defs"
            / "basic_drainage_all_bits.inp"
        )

        # Manually convert to ha to make comparable to synthetic_write
        data_dict["subs"]["area"] /= 10000

        # Write the model with data_dict_to_inp
        stt.data_dict_to_inp(
            stt.format_to_swmm_dict(**data_dict), template_fid, comparison_file
        )

        # Compare
        new_input_file = addresses.model_paths.inp
        are_files_identical = filecmp.cmp(
            new_input_file, comparison_file, shallow=False
        )
        if not are_files_identical:
            with new_input_file.open("r") as file1, comparison_file.open("r") as file2:
                diff = difflib.unified_diff(
                    file1.readlines(),
                    file2.readlines(),
                    fromfile=str(new_input_file),
                    tofile=str(comparison_file),
                )
            print("".join(diff))
        assert are_files_identical, "The files are not identical"

        # Test that it doesn't break there are more outfalls than links
        nodes.loc["new_node"] = nodes.iloc[0].copy()
        nodes.reset_index().to_file(addresses.model_paths.nodes)

        # Check that there will be more outfalls than edges
        assert sum(~nodes.index.astype(str).isin(edges.u.astype(str))) > edges.shape[0]

        # Run to check
        stt.synthetic_write(addresses)


def test_format_to_swmm_dict():
    """Test the format_format_to_swmm_dict function.

    Writes a formatted dict to a model and checks that it runs without
    error.
    """
    data_dict = generate_data_dict()
    data_dict = stt.format_to_swmm_dict(**data_dict)

    rain_fid = (
        Path(__file__).parent.parent / "src" / "swmmanywhere" / "defs" / "storm.dat"
    )
    with tempfile.TemporaryDirectory(dir=".") as temp_dir:
        temp_dir = Path(temp_dir)
        shutil.copy(rain_fid, temp_dir / "storm.dat")
        temp_fid = temp_dir / "temp.inp"
        stt.data_dict_to_inp(data_dict, fid, temp_fid)
        with pyswmm.Simulation(str(temp_fid)) as sim:
            for ind, step in enumerate(sim):
                pass



"""Post processing module for SWMManywhere.

A module containing functions to format and write processed data into SWMM .inp
files.
"""

from __future__ import annotations

import re
import shutil
from pathlib import Path
from typing import Any, Literal

import geopandas as gpd
import numpy as np
import pandas as pd
import yaml

from swmmanywhere.filepaths import FilePaths
from swmmanywhere.logging import logger


def synthetic_write(addresses: FilePaths):
    """Load synthetic data and write to SWMM input file.

    Loads nodes, edges and subcatchments from synthetic data, assumes that
    these are all located in `addresses`. Fills in appropriate default values
    for many SWMM parameters. More parameters are available to edit (see
    defs/swmm_conversion.yml). Identifies outfalls and automatically ensures
    that they have only one link to them (as is required by SWMM). Formats
    (with format_to_swmm_dict) and writes (with data_dict_to_inp) the data to
    a SWMM input (.inp) file.

    Args:
        addresses (FilePaths): A dictionary of file paths.
    """
    nodes = gpd.read_file(addresses.model_paths.nodes)
    edges = gpd.read_file(addresses.model_paths.edges)

    if addresses.model_paths.subcatchments.suffix == ".geoparquet":
        subs = gpd.read_parquet(addresses.model_paths.subcatchments)
    else:
        subs = gpd.read_file(addresses.model_paths.subcatchments)
    subs = subs.loc[subs.id.isin(nodes.id)]

    # Extract SWMM relevant data
    edges = edges[["id", "u", "v", "diameter", "length"]]
    nodes = nodes[["id", "x", "y", "chamber_floor_elevation", "surface_elevation"]]
    subs = subs[["id", "geometry", "area", "slope", "width", "rc"]]

    # Nodes
    nodes["id"] = nodes["id"].astype(str)
    nodes["max_depth"] = nodes.surface_elevation - nodes.chamber_floor_elevation
    nodes["surcharge_depth"] = 0
    nodes["flooded_area"] = 100  # TODO arbitrary... not sure how to calc this
    nodes["manhole_area"] = 0.5

    # Subs
    subs["id"] = subs["id"].astype(str)
    subs["subcatchment"] = subs["id"] + "-sub"
    subs["rain_gage"] = 1  # TODO revise when revising storms
    subs["area"] /= 10000  # convert to ha

    # Edges
    edges["u"] = edges["u"].astype(str)
    edges["v"] = edges["v"].astype(str)
    edges["roughness"] = 0.01
    edges["capacity"] = 1e10  # capacity in swmm is a hard limit

    # Outfalls (create new nodes that link to the stores connected tothem
    outfalls = nodes.loc[~nodes.id.isin(edges.u)].copy()
    outfalls["id"] = outfalls["id"] + "_outfall"

    # Reduce elevation to ensure flow
    outfalls["chamber_floor_elevation"] -= 5
    outfalls["x"] -= 50
    outfalls["y"] -= 50

    # Link stores to outfalls
    new_edges = pd.DataFrame(columns=edges.columns, index=range(len(outfalls)))
    new_edges["u"] = outfalls["id"].str.replace("_outfall", "").values
    new_edges["v"] = outfalls["id"].values
    new_edges["id"] = [f"{u}-{v}" for u, v in zip(new_edges["u"], new_edges["v"])]
    new_edges["diameter"] = edges.diameter.max()
    new_edges["length"] = (50**2 + 50**2) ** 0.5
    new_edges["roughness"] = 0.01
    new_edges["capacity"] = 1e10

    # Append new edges
    edges = pd.concat([edges, new_edges], ignore_index=True)

    # Create event
    # TODO will need some updating if multiple rain gages
    # TODO automatically match units to storm.csv?
    event = {
        "name": "1",
        "unit": "mm",
        "interval": "00:05",  # hh:mm
        "fid": str(addresses.bbox_paths.precipitation),
    }

    # Locate raingage(s) on the map
    symbol = {
        "x": nodes.x.min(),
        "y": nodes.y.min(),
        "name": "1",  # matches event name(s)
    }

    # Template SWMM input file
    existing_input_file = Path(__file__).parent / "defs" / "basic_drainage_all_bits.inp"

    # Format to dict
    data_dict = format_to_swmm_dict(nodes, outfalls, edges, subs, event, symbol)

    # Write new input file
    data_dict_to_inp(data_dict, existing_input_file, addresses.model_paths.inp)


def overwrite_section(data: np.ndarray, section: str, fid: Path):
    """Overwrite a section of a SWMM .inp file with new data.

    Args:
        data (np.ndarray): Data array to be written to the SWMM .inp file.
        section (str): Section of the SWMM .inp file to be overwritten.
        fid (Path): File path to the SWMM .inp file.

    Example:
        data = np.array([
                ['1', '1', '1', '1.166', '100', '500', '0.5', '0', 'empty'],
                ['2', '1', '1', '1.1', '100', '500', '0.5', '0', 'empty'],
                ['3', '1', '1', '2', '100', '400', '0.5', '0', 'empty']])
        fid = 'my_pre_existing_swmm_input_file.inp'
        section = '[SUBCATCHMENTS]'
        overwrite_section(data, section, fid)
    """
    # Read the existing SWMM .inp file
    with fid.open("r") as infile:
        lines = infile.readlines()

    # Create a flag to indicate whether we are within the target section
    within_target_section = False

    # Iterate through the lines and make modifications as needed
    with fid.open("w") as outfile:
        for ix, line in enumerate(lines):
            if line.strip() != section and re.search(r"\[.*?\]", line):
                within_target_section = False

            if line.strip() != section and not within_target_section:
                outfile.write(line)  # Write lines outside the target section

            if line.strip() != section:
                continue

            within_target_section = True
            outfile.write(line)  # Write the start section header

            # Write headers
            i = 1
            while lines[ix + i][0] == ";":
                outfile.write(lines[ix + i])  # Write column headers
                i += 1

            example_line = lines[ix + i]
            # print('example_line {1}: {0}'.format(
            #     example_line.replace('\n', ''), section))
            # print('note - this line must have at least as many column')
            # print('entries as all other rows in this section\n')
            pattern = r"(\s+)"

            # Find all matches of the pattern in the input line
            matches = re.findall(pattern, example_line)

            # Calculate the space counts by taking the length of each match
            space_counts = [
                len(x) + len(y) for x, y in zip(matches, example_line.split())
            ]
            if not space_counts:
                if data.shape[0] != 0:
                    logger.warning("no template for data?")
                continue

            space_counts[-1] -= 1
            new_text = ""
            if data is None:
                continue

            for i, row in enumerate(data):
                if section == "[CONTROLS]":
                    new_text = row
                    continue

                formatted_row = []
                for x, y in zip(row, space_counts):
                    formatted_value = "{0:<{1}}".format(x, max(y, len(str(x)) + 1))
                    formatted_row.append(formatted_value)
                new_line = "{0}\n".format("".join(formatted_row))
                new_text += new_line

            outfile.write(new_text)  # Write the new content
            outfile.write("\n")


def change_flow_routing(
    routing_method: Literal["KINWAVE", "DYNWAVE", "STEADY"], file_path: Path
) -> None:
    """Replace the flow routing method in a SWMM inp file with a new method, in-place.

    Args:
        file_path : Path
            Path to the SWMM inp file to be modified.
        routing_method : {"KINWAVE", "DYNWAVE", "STEADY"}
            The new flow routing method to be used. Available options are:
            ``KINWAVE``, ``DYNWAVE``, or ``STEADY``.
    """
    if routing_method.upper() not in ("KINWAVE", "DYNWAVE", "STEADY"):
        raise ValueError(
            "routing_method must be one of 'KINWAVE', 'DYNWAVE', or 'STEADY'."
        )
    updated_contents = re.sub(
        r"^FLOW_ROUTING\s+.*$",
        f"FLOW_ROUTING {routing_method.upper()}",
        file_path.read_text(),
        flags=re.MULTILINE,
    )
    file_path.write_text(updated_contents)


def data_dict_to_inp(
    data_dict: dict[str, np.ndarray],
    base_input_file: Path,
    new_input_file: Path,
    routing: Literal["KINWAVE", "DYNWAVE", "STEADY"] = "DYNWAVE",
):
    """Write a SWMM .inp file from a dictionary of data arrays.

    Args:
        data_dict (dict[str, np.ndarray]): Dictionary of data arrays. Where
            each key is a SWMM section and each value is a numpy array of
            data to be written to that section. The existing section is
            overwritten
        base_input_file (Path): File path to the example/template .inp file.
        new_input_file (Path): File path to the new SWMM .inp file.
        routing (str, optional): Flow routing method (KINWAVE, DYNWAVE,
            STEADY). Defaults to "DYNWAVE".
    """
    shutil.copy2(base_input_file, new_input_file)

    # Write the inp file
    for key, data in data_dict.items():
        start_section = "[{0}]".format(key)

        overwrite_section(data, start_section, new_input_file)

    # Set the flow routing
    change_flow_routing(routing, new_input_file)


def explode_polygon(row: pd.Series):
    """Explode a polygon into a DataFrame of coordinates.

    Args:
        row (pd.Series): A row of a GeoDataFrame containing a polygon.

    Example:
        >>> import geopandas as gpd
        >>> from shapely.geometry import Polygon
        >>> df = pd.Series({'subcatchment' : '1',
        ...                 'geometry' : Polygon([(0,0), (1,0),
        ...                                       (1,1), (0,1)])})
        >>> explode_polygon(df)
             x    y subcatchment
        0  0.0  0.0            1
        1  1.0  0.0            1
        2  1.0  1.0            1
        3  0.0  1.0            1
        4  0.0  0.0            1
    """
    # Get the vertices of the polygon
    vertices = list(row["geometry"].exterior.coords)

    # Create a new DataFrame for this row
    df = pd.DataFrame(columns=["x", "y"], data=vertices)
    df["subcatchment"] = row["subcatchment"]
    return df


def format_to_swmm_dict(
    nodes: pd.DataFrame,
    outfalls: pd.DataFrame,
    conduits: pd.DataFrame,
    subs: gpd.GeoDataFrame,
    event: dict[str, Any],
    symbol: dict[str, Any],
) -> dict[str, np.ndarray]:
    """Format data to a dictionary of data arrays with columns matching SWMM.

    These data are the parameters of all assets that are written to the SWMM
    input file. More parameters are available to edit (see
    defs/swmm_conversion.yml).

    Args:
        nodes (pd.DataFrame): GeoDataFrame of nodes. With at least columns:
            'id', 'x', 'y', 'max_depth', 'chamber_floor_elevation',
            'manhole_area'.
        outfalls (pd.DataFrame): GeoDataFrame of outfalls. With at least
            columns: 'id', 'chamber_floor_elevation'.
        conduits (pd.DataFrame): GeoDataFrame of conduits. With at least
            columns: 'id', 'u', 'v', 'length', 'roughness', 'shape_swmm',
            'diameter'.
        subs (gpd.GeoDataFrame): GeoDataFrame of subcatchments. With at least
            columns: 'subcatchment', 'rain_gage', 'id', 'area', 'rc', 'width',
            'geometry',
        event (dict): Dict describing storm event. With at least
            keys: 'name', 'unit', 'interval', 'fid'.
        symbol (dict): Dict with coordinates of rain gage. With at least keys:
            'x', 'y', 'name'.

    Example:
        >>> import os
        >>> import geopandas as gpd
        >>> from shapely.geometry import Point, Polygon
        >>> nodes = gpd.GeoDataFrame({'id' : ['node1', 'node2'],
        ...                        'x' : [0, 1],
        ...                        'y' : [0, 1],
        ...                        'max_depth' : [1, 1],
        ...                        'chamber_floor_elevation' : [1, 1],
        ...                        'manhole_area' : [1,1]
        ...                        })
        >>> outfalls = gpd.GeoDataFrame({'id' : ['outfall3'],
        ...                                'chamber_floor_elevation' : [1],
        ...                                'x' : [0],
        ...                                'y' : [0]})
        >>> conduits = gpd.GeoDataFrame({'id' : ['link1','link2'],
        ...                                'u' : ['node1','node2'],
        ...                                'v' : ['node2','outfall3'],
        ...                                'length' : [1,1],
        ...                                'roughness' : [1,1],
        ...                                'shape_swmm' : ['CIRCULAR','CIRCULAR'],
        ...                                'diameter' : [1,1],
        ...                                'capacity' : [0.1,0.1]
        ...                                })
        >>> subs = gpd.GeoDataFrame({'subcatchment' : ['sub1'],
        ...                                'rain_gage' : ['1'],
        ...                                'id' : ['node1'],
        ...                                'area' : [1],
        ...                                'rc' : [1],
        ...                                'width' : [1],
        ...                                'slope' : [0.001],
        ...                                'geometry' : [Polygon([(0.0,0.0),
        ...                                                       (1.0,0.0),
        ...                                                       (1.0,1.0),
        ...                                                       (0.0,1.0)])]})
        >>> rain_fid = os.path.join(os.path.dirname(os.path.abspath(__file__)),
        ...                '..',
        ...                    'swmmanywhere',
        ...                    'defs',
        ...                    'storm.dat')
        >>> event = {'name' : '1',
        ...                'unit' : 'mm',
        ...                'interval' : 1,
        ...                'fid' : rain_fid}
        >>> symbol = {'x' : 0,
        ...            'y' : 0,
        ...            'name' : 'name'}
        >>> data_dict = format_to_swmm_dict(nodes,
        ...                                    outfalls,
        ...                                    conduits,
        ...                                    subs,
        ...                                    event,
        ...                                    symbol)
    """
    # Get the directory of the current module
    current_dir = Path(__file__).parent

    # TODO use 'load_yaml_from_defs'
    # Create the path to iso_converter.yml
    iso_path = current_dir / "defs" / "swmm_conversion.yml"

    # Load conversion mapping from YAML file
    with iso_path.open("r") as file:
        conversion_dict = yaml.safe_load(file)

    ## Create nodes, coordinates and map dimensions
    dims = {
        "x1": nodes.x.min(),
        "y1": nodes.y.min(),
        "x2": nodes.x.max(),
        "y2": nodes.y.max(),
    }
    dims = {x: str(y) + " " for x, y in dims.items()}

    map_dimensions = pd.Series(dims).reset_index().set_index("index").T
    polygons = subs[["subcatchment", "geometry"]].copy()

    # Format dicts to DataFrames
    event = pd.Series(event).reset_index().set_index("index").T
    symbol = pd.Series(symbol).reset_index().set_index("index").T

    # Apply the function to each row
    polygons = pd.concat(polygons.apply(explode_polygon, axis=1).tolist())

    ## Specify sections
    shps = {
        "SUBCATCHMENTS": subs,
        "CONDUITS": conduits,
        "OUTFALLS": outfalls,
        "STORAGE": nodes,
        "XSECTIONS": conduits,
        "SUBAREAS": subs,
        "INFILTRATION": subs,
        "COORDINATES": pd.concat([nodes, outfalls], axis=0),
        "MAP": map_dimensions,
        "Polygons": polygons,
        "PUMPS": None,
        "ORIFICES": None,
        "WEIRS": None,
        "OUTLETS": None,
        "JUNCTIONS": None,
        "RAINGAGES": event,
        "SYMBOLS": symbol,
    }

    # Fill backslash columns and store data in data_dict in the correct order
    import numpy.typing as npt

    def _fill_backslash_columns(
        shp: pd.DataFrame | None, key: str
    ) -> npt.ArrayLike | None:
        if shp is None:
            return None

        # Extract SWMM order and default values
        columns = conversion_dict[key]["iwcolumns"]
        numeric_cols = shp.select_dtypes(include=[np.number]).columns
        shp[numeric_cols] = shp[numeric_cols].fillna(0)

        # Find columns with a default specified
        cols_default = [c[1:] for c in columns if c.startswith("/")]

        # Fill columns with defaults
        shp[["/" + c for c in cols_default]] = np.array(cols_default).T
        return shp[columns].values

    data_dict = {key: _fill_backslash_columns(shp, key) for key, shp in shps.items()}
    return data_dict



"""Load example data."""

from __future__ import annotations

from pathlib import Path

from swmmanywhere.graph_utilities import load_graph

demo_graph = load_graph(Path(__file__).parent / "demo_graph.json")



"""Metric utilities module for SWMManywhere.

A module for metrics, the metrics registry object and utilities for calculating
metrics (such as NSE or timeseries data alignment) used in SWMManywhere.
"""

from __future__ import annotations

from collections import defaultdict
from itertools import product
from typing import Callable, Optional, get_type_hints

import cytoolz.curried as tlz
import geopandas as gpd
import joblib
import networkx as nx
import numpy as np
import pandas as pd
import shapely
from scipy import stats

import netcomp
from swmmanywhere.logging import logger
from swmmanywhere.parameters import MetricEvaluation


class MetricRegistry(dict):
    """Registry object."""

    def _log_completion(self, func):
        def _wrapper(*args, **kwargs):
            result = func(*args, **kwargs)
            logger.info(f"{func.__name__} completed")
            return result

        return _wrapper

    def register(self, func: Callable) -> Callable:
        """Register a metric."""
        if func.__name__ in self:
            raise ValueError(f"{func.__name__} already in the metric registry!")

        allowable_params = {
            "synthetic_results": pd.DataFrame,
            "real_results": pd.DataFrame,
            "synthetic_subs": gpd.GeoDataFrame,
            "real_subs": gpd.GeoDataFrame,
            "synthetic_G": nx.Graph,
            "real_G": nx.Graph,
            "metric_evaluation": MetricEvaluation,
        }

        # Use get_type_hints to resolve annotations,
        # considering 'from __future__ import annotations'
        type_hints = get_type_hints(func)

        for param, annotation in type_hints.items():
            if param in ("kwargs", "return"):
                continue
            if param not in allowable_params:
                raise ValueError(f"{param} of {func.__name__} not allowed.")
            if annotation != allowable_params[param]:
                raise ValueError(
                    f"""{param} of {func.__name__} should be of
                                 type {allowable_params[param]}, not 
                                 {annotation}."""
                )
        self[func.__name__] = self._log_completion(func)
        return func

    def __getattr__(self, name):
        """Get a metric from the graphfcn dict."""
        try:
            return self[name]
        except KeyError:
            raise AttributeError(f"{name} NOT in the metric registry!")


metrics = MetricRegistry()


def validate_metric_list(metric_list: list[str]) -> None:
    """Validate a list of metrics.

    Validate that all metrics in the metric list are registered.

    Args:
        metric_list (list[str]): A list of metrics to validate.

    Raises:
        ValueError: If a metric is not registered.
    """
    not_exists = [m for m in metric_list if m not in metrics]
    if not_exists:
        raise ValueError(f"Metrics are not registered:\n{', '.join(not_exists)}")


def apply_warmup(results: pd.DataFrame, warmup: float) -> pd.DataFrame:
    """Apply warmup to results.

    Apply a warmup period to the results dataframe. Removes all data during the
    warmup period.


    Args:
        results (pd.DataFrame): The results dataframe.
        warmup (float): The warmup period as a fraction of the total time period.

    Returns:
        pd.DataFrame: The results dataframe with the warmup period applied.
    """
    return results.loc[
        results.date
        >= (results.date.min() + warmup * (results.date.max() - results.date.min()))
    ]


def iterate_metrics(
    synthetic_results: pd.DataFrame | None = None,
    synthetic_subs: gpd.GeoDataFrame | None = None,
    synthetic_G: nx.Graph | None = None,
    real_results: pd.DataFrame | None = None,
    real_subs: gpd.GeoDataFrame | None = None,
    real_G: nx.Graph | None = None,
    metric_list: list[str] | None = None,
    metric_evaluation: MetricEvaluation | None = None,
) -> dict[str, float]:
    """Iterate a list of metrics over a graph.

    Args:
        synthetic_results (pd.DataFrame | None): The synthetic results.
        synthetic_subs (gpd.GeoDataFrame | None): The synthetic subcatchments.
        synthetic_G (nx.Graph | None): The synthetic graph.
        real_results (pd.DataFrame | None): The real results.
        real_subs (gpd.GeoDataFrame | None): The real subcatchments.
        real_G (nx.Graph | None): The real graph.
        metric_list (list[str] | None): A list of metrics to iterate.
        metric_evaluation (MetricEvaluation | None): The metric evaluation parameters.

    Returns:
        dict[str, float]: The results of the metrics.
    """
    if metric_list is None:
        return {}

    validate_metric_list(metric_list)

    if (metric_evaluation is not None) and (synthetic_results is not None):
        synthetic_results = apply_warmup(synthetic_results, metric_evaluation.warmup)

    kwargs = {
        "synthetic_results": synthetic_results,
        "synthetic_subs": synthetic_subs,
        "synthetic_G": synthetic_G,
        "real_results": real_results,
        "real_subs": real_subs,
        "real_G": real_G,
        "metric_evaluation": metric_evaluation,
    }

    return {m: metrics[m](**kwargs) for m in metric_list}


def extract_var(df: pd.DataFrame, var: str) -> pd.DataFrame:
    """Extract var from a dataframe."""
    df_ = df.loc[df.variable == var].copy()
    df_.loc[:, "duration"] = (df_.date - df_.date.min()).dt.total_seconds()
    return df_


# Restriction registry
restriction_registry = {}


def register_restriction(restriction_func: Callable):
    """Register a restriction function.

    Register a restriction function to the restriction_registry. A restriction
    allows for the restriction of certain combinations of variables within the
    metric_factory. The function should take three arguments, 'scale', 'metric',
    and 'variable', and should raise a ValueError if the combination is not
    allowed. The function should be registered with the '@register_restriction'.

    Args:
        restriction_func (Callable): The restriction function to register.
    """
    name = restriction_func.__name__

    # Check if the function is already registered
    if name in restriction_registry:
        raise ValueError(f"Restriction function '{name}' already registered.")

    # Validate the restriction
    args = list(get_type_hints(restriction_func).keys())
    if args != ["scale", "metric", "variable"]:
        raise ValueError(
            f"""Restriction {restriction_func.__name__} requires 
                         args ('scale', 'metric', 'variable')."""
        )

    # Add the function to the registry
    restriction_registry[name] = restriction_func
    return restriction_func


@register_restriction
def restriction_on_scale(scale: str, metric: str, variable: str):
    """Restriction on scale.

    Restrict the design variables to the outfall scale if the metric is 'relerror'.

    Args:
        scale (str): The scale of the metric.
        metric (str): The metric.
        variable (str): The variable.
    """
    if variable in ("length", "nmanholes", "npipes") and scale != "outfall":
        raise ValueError(f"Variable {variable} only supported at the outfall scale")


@register_restriction
def restriction_on_metric(scale: str, metric: str, variable: str):
    """Restriction on metric.

    Restrict the design variables to use 'relerror' only.

    Args:
        scale (str): The scale of the metric.
        metric (str): The metric.
        variable (str): The variable.
    """
    if variable in ("length", "nmanholes", "npipes") and metric != "relerror":
        raise ValueError(f"Variable {variable} only valid with relerror metric")


# Coefficient Registry
coef_registry = {}


def register_coef(coef_func: Callable):
    """Register a coefficient function.

    Register a coefficient function to the coef_registry. The function should
    take two arguments, 'y' and 'yhat', and return a float. The function should
    be registered with the '@register_coef' decorator.

    Args:
        coef_func (Callable): The coefficient function to register.
    """
    name = coef_func.__name__

    # Check if the function is already registered
    if name in coef_registry:
        raise ValueError(f"Coefficient function '{name}' already registered.")

    # Validate the function
    args = list(get_type_hints(coef_func).keys())
    if "y" != args[0] or "yhat" != args[1]:
        raise ValueError(f"Coef {coef_func.__name__} requires args ('y', 'yhat').")

    # Add the function to the registry
    coef_registry[name] = coef_func
    return coef_func


@register_coef
def relerror(y: np.ndarray, yhat: np.ndarray) -> float:
    r"""Relative error, relerror.

    Calculate the relative error:

    $$
    relerror = \frac{mean(synthetic) - mean(real)}
                 {mean(real)}
    $$

    where:

    - \(synthetic\) is the synthetic data,
    - \(real\) is the real data,


    Args:
        y (np.ndarray): The real data.
        yhat (np.ndarray): The synthetic data.

    Returns:
        float: The relerror value.
    """
    total_observed = y.mean()
    if total_observed == 0:
        return np.inf
    return (yhat.mean() - total_observed) / total_observed


@register_coef
def nse(y: np.ndarray, yhat: np.ndarray) -> float:
    r"""Calculate Nash-Sutcliffe efficiency (NSE).

    Calculate the Nash-Sutcliffe efficiency (NSE):

    $$
    NSE = 1 - \frac{\sum_{i=1}^{n} (Q_{obs,i} - Q_{sim,i})^2}
                   {\sum_{i=1}^{n} (Q_{obs,i} - \overline{Q}_{obs})^2}
    $$

    where:

    - $Q_{obs,i}$ is the observed value at time $i$,
    - $Q_{sim,i}$ is the simulated value at time $i$,
    - $\overline{Q}_{obs}$ is the mean observed value over the simulation period,
    - $n$ is the number of time steps in the simulation period.

    Args:
        y (np.array): Observed data array.
        yhat (np.array): Simulated data array.

    Returns:
        float: The NSE value.
    """
    if np.std(y) == 0:
        return np.inf
    return 1 - np.sum(np.square(y - yhat)) / np.sum(np.square(y - np.mean(y)))


@register_coef
def kge(y: np.ndarray, yhat: np.ndarray) -> float:
    r"""Calculate the Kling-Gupta Efficiency (KGE) between simulated and observed data.

    Calculate KGE with the 2009 formulation:
    $$
    KGE = 1 -
        \sqrt{ (r - 1)^2 +
               (\frac{\sigma_{sim}}{\sigma_{obs}} - 1)^2 +
               (\frac{\mu_{sim}}{\mu_{obs}} - 1)^2
        }
    $$

    where:

    - $r$ is the correlation coefficient between observed and simulated value,
    - $\sigma_{sim}$ and $\sigma_{obs}$ are the standard deviations of the
        simulated and observed value, respectively,
    - $\mu_{sim}$ and $\mu_{obs}$ are the means of the simulated and observed
        value, respectively.


    Args:
        y (np.array): Observed data array.
        yhat (np.array): Simulated data array.

    Returns:
        float: The KGE value.
    """
    if (np.std(y) == 0) | (np.mean(y) == 0):
        return np.inf
    if np.std(yhat) == 0:
        r = 0
    else:
        r = np.corrcoef(yhat, y)[0, 1]
    alpha = np.std(yhat) / np.std(y)
    beta = np.mean(yhat) / np.mean(y)
    kge = 1 - np.sqrt((r - 1) ** 2 + (alpha - 1) ** 2 + (beta - 1) ** 2)
    return kge


def align_by_id(
    synthetic_results: pd.DataFrame,
    real_results: pd.DataFrame,
    variable: str,
    syn_ids: list,
    real_ids: list,
):
    """Align and interpolate data by id.

    Aggregate synthetic and real results by date for specifics ids (i.e., sum
    up over all ids - so we are only comparing timeseries for one aggregation).
    Align the synthetic and real dates. In cases where the synthetic
    data is does not overlap the real data, the value is interpolated.

    Args:
        synthetic_results (pd.DataFrame): The synthetic results.
        real_results (pd.DataFrame): The real results.
        variable (str): The variable to align and calculate coef_func for.
        syn_ids (list): The ids of the synthetic data to subselect for.
        real_ids (list): The ids of the real data to subselect for.
        coef_func (Callable, optional): The coefficient to calculate.
            Defaults to nse.

    Returns:
        pd.DataFrame: The aligned and interpolated data.
    """
    synthetic_results = synthetic_results.copy()
    real_results = real_results.copy()

    # Format dates
    synthetic_results["date"] = pd.to_datetime(synthetic_results["date"])
    real_results["date"] = pd.to_datetime(real_results["date"])

    # Help alignment
    synthetic_results["id"] = synthetic_results["id"].astype(str)
    real_results["id"] = real_results["id"].astype(str)
    syn_ids = [str(x) for x in syn_ids]
    real_ids = [str(x) for x in real_ids]

    # Extract data
    syn_data = extract_var(synthetic_results, variable)
    syn_data = syn_data.loc[syn_data["id"].isin(syn_ids)]
    syn_data = syn_data.groupby("date").value.sum()

    real_data = extract_var(real_results, variable)
    real_data = real_data.loc[real_data["id"].isin(real_ids)]
    real_data = real_data.groupby("date").value.sum()

    # Align data
    df = pd.merge(
        syn_data,
        real_data,
        left_index=True,
        right_index=True,
        suffixes=("_syn", "_real"),
        how="outer",
    ).sort_index()

    # Interpolate to time in real data
    df["value_syn"] = df.value_syn.interpolate().to_numpy()
    df = df.dropna(subset=["value_real"])

    return df


def create_subgraph(G: nx.Graph, nodes: list) -> nx.Graph:
    """Create a subgraph.

    Create a subgraph of G based on the nodes list. Taken from networkx
    documentation: https://networkx.org/documentation/stable/reference/classes/generated/networkx.Graph.subgraph.html

    Args:
        G (nx.Graph): The original graph.
        nodes (list): The list of nodes to include in the subgraph.

    Returns:
        nx.Graph: The subgraph.
    """
    # Create a subgraph SG based on a (possibly multigraph) G
    SG = G.__class__()
    SG.add_nodes_from((n, G.nodes[n]) for n in nodes)
    if SG.is_multigraph():
        SG.add_edges_from(
            (n, nbr, key, d)
            for n, nbrs in G.adj.items()
            if n in nodes
            for nbr, keydict in nbrs.items()
            if nbr in nodes
            for key, d in keydict.items()
        )
    else:
        SG.add_edges_from(
            (n, nbr, d)
            for n, nbrs in G.adj.items()
            if n in nodes
            for nbr, d in nbrs.items()
            if nbr in nodes
        )
    SG.graph.update(G.graph)
    return SG


def median_coef_by_group(
    results: pd.DataFrame, gb_key: str, coef_func: Callable = nse
) -> float:
    """Median coef_func value by group.

    Calculate the median coef_func value of a variable over time
    for each group in the results dataframe, and return the median of these
    values. Assumes that the results dataframe has a 'value_real' and 'value_syn'
    and that these properly line up.

    Args:
        results (pd.DataFrame): The results dataframe.
        gb_key (str): The column to group by.
        coef_func (Callable): The coefficient to calculate. Default is nse.

    Returns:
        float: The median coef_func value.
    """
    val = results.groupby(gb_key).apply(lambda x: coef_func(x.value_real, x.value_syn))
    val = val[np.isfinite(val)]
    return val.median()


def nodes_to_subs(G: nx.Graph, subs: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
    """Nodes to subcatchments.

    Classify the nodes of the graph to the subcatchments of the subs dataframe.

    Args:
        G (nx.Graph): The graph.
        subs (gpd.GeoDataFrame): The subcatchments.

    Returns:
        gpd.GeoDataFrame: A dataframe from the nodes and data, and the
            subcatchment information, distinguished by the column 'sub_id'.
    """
    nodes_df = pd.DataFrame([{"id": x, **d} for x, d in G.nodes(data=True)])
    nodes_joined = gpd.GeoDataFrame(
        nodes_df,
        geometry=gpd.points_from_xy(nodes_df.x, nodes_df.y),
        crs=G.graph["crs"],
    ).sjoin(subs.rename(columns={"id": "sub_id"}), how="inner", predicate="within")
    return nodes_joined


def best_outfall_match(
    synthetic_G: nx.Graph, real_subs: gpd.GeoDataFrame
) -> tuple[nx.Graph, int | None]:
    """Best outfall match.

    Identify the outfall with the most nodes within the real_subs and return the
    subgraph of the synthetic graph of nodes that drain to that outfall.

    Args:
        synthetic_G (nx.Graph): The synthetic graph.
        real_subs (gpd.GeoDataFrame): The real subcatchments.

    Returns:
        nx.Graph: The subgraph of the synthetic graph for the outfall with the
            most nodes within the real_subs. Empty if no match is made.
        int: The id of the outfall. None if no outfall is found.
    """
    nodes_joined = nodes_to_subs(synthetic_G, real_subs)

    if nodes_joined.shape[0] == 0:
        return (nx.Graph(), None)

    # Select the most common outfall
    outfall = nodes_joined.outfall.value_counts().idxmax()

    # Subselect the matching graph
    outfall_nodes = [
        n for n, d in synthetic_G.nodes(data=True) if d.get("outfall", None) == outfall
    ]
    sg = create_subgraph(synthetic_G, outfall_nodes)
    return sg, outfall


def dominant_outfall(G: nx.DiGraph, results: pd.DataFrame) -> tuple[nx.DiGraph, int]:
    """Dominant outfall.

    Identify the outfall with highest flow along it and return the
    subgraph of the graph of nodes that drain to that outfall.

    Args:
        G (nx.DiGraph): The graph.
        results (pd.DataFrame): The results, which include a 'flow' and 'id'
            column.

    Returns:
        nx.Graph: The subgraph of nodes/arcs that the reach max flow outfall
        int: The id of the outfall.
    """
    # Identify outfalls of the graph
    outfalls = [n for n, outdegree in G.out_degree() if outdegree == 0]
    outfall_arcs = [d["id"] for u, v, d in G.edges(data=True) if v in outfalls]

    # Identify the outfall with the highest flow
    outfall_flows = results.loc[
        (results.variable == "flow") & (results["id"].isin(outfall_arcs))
    ]
    max_outfall_arc = outfall_flows.groupby("id").value.median().idxmax()
    max_outfall = [v for u, v, d in G.edges(data=True) if d["id"] == max_outfall_arc][0]

    # Subselect the matching graph
    sg = create_subgraph(G, nx.ancestors(G, max_outfall) | {max_outfall})
    return sg, max_outfall


def nc_compare(G1, G2, funcname, **kw):
    """Compare two graphs using netcomp."""
    A1, A2 = [nx.adjacency_matrix(G) for G in (G1, G2)]
    return getattr(netcomp, funcname)(A1, A2, **kw)


def edge_betweenness_centrality(
    G: nx.Graph,
    normalized: bool = True,
    weight: Optional[str] = "weight",
    njobs: int = -1,
):
    """Parallel betweenness centrality function."""
    njobs = 1  # joblib.cpu_count(True) if njobs == -1 else njobs #TODO hotfix
    node_chunks = tlz.partition_all(G.order() // njobs, G.nodes())
    bt_func = tlz.partial(
        nx.edge_betweenness_centrality_subset, G=G, normalized=normalized, weight=weight
    )
    bt_sc = joblib.Parallel(n_jobs=njobs)(
        joblib.delayed(bt_func)(sources=nodes, targets=G.nodes())
        for nodes in node_chunks
    )

    # Merge the betweenness centrality results
    bt_c: dict[int, float] = defaultdict(float)
    for bt in bt_sc:
        for n, v in bt.items():
            bt_c[n] += v
    return bt_c


def align_by_shape(
    var,
    synthetic_results: pd.DataFrame,
    real_results: pd.DataFrame,
    shapes: gpd.GeoDataFrame,
    synthetic_G: nx.Graph,
    real_G: nx.Graph,
    key: str = "sub_id",
) -> pd.DataFrame:
    """Align by subcatchment.

    Align synthetic and real results by shape and return the results. If multiple
    ids exist in the same shape, these are aggregated by sum.

    Args:
        var (str): The variable to align.
        synthetic_results (pd.DataFrame): The synthetic results.
        real_results (pd.DataFrame): The real results.
        shapes (gpd.GeoDataFrame): The shapes to align by (e.g., grid or real_subs).
        synthetic_G (nx.Graph): The synthetic graph.
        real_G (nx.Graph): The real graph.
        key (str): The column to align by.
    """
    synthetic_joined = nodes_to_subs(synthetic_G, shapes)
    real_joined = nodes_to_subs(real_G, shapes)

    # Extract data
    real_results = extract_var(real_results, var)
    synthetic_results = extract_var(synthetic_results, var)

    # Format to help alignment
    real_results["id"] = real_results["id"].astype(str)
    synthetic_results["id"] = synthetic_results["id"].astype(str)
    real_joined["id"] = real_joined["id"].astype(str)
    synthetic_joined["id"] = synthetic_joined["id"].astype(str)

    # Align data
    synthetic_results = pd.merge(
        synthetic_results, synthetic_joined[["id", key]], on="id"
    )
    synthetic_gb = synthetic_results.groupby(["date", key]).value.sum().reset_index()
    real_results = pd.merge(real_results, real_joined[["id", key]], on="id")
    real_gb = real_results.groupby(["date", key]).value.sum().reset_index()
    results = pd.merge(
        real_gb[["date", key, "value"]],
        synthetic_gb[["date", key, "value"]],
        on=["date", key],
        suffixes=("_real", "_syn"),
        how="outer",
    )

    syn_interp = (
        results.groupby(key)
        .apply(func=lambda x: x.set_index("date")[["value_syn"]].interpolate())
        .reset_index()
    )
    results = pd.merge(
        results.drop("value_syn", axis=1), syn_interp, on=["sub_id", "date"]
    )
    results = results.dropna(subset=["value_real"])

    return results


def create_grid(bbox: tuple, scale: float | tuple[float, float]) -> gpd.GeoDataFrame:
    """Create a grid of polygons.

    Create a grid of polygons based on the bounding box and scale.

    Args:
        bbox (tuple): The bounding box coordinates in the format (minx, miny,
            maxx, maxy).
        scale (float | tuple): The scale of the grid. If a tuple, the scale is
            (dx, dy). Otherwise, the scale is dx = dy = scale.

    Returns:
        gpd.GeoDataFrame: A geodataframe of the grid.
    """
    minx, miny, maxx, maxy = bbox

    if isinstance(scale, tuple):
        if len(scale) != 2:
            raise ValueError(
                f"""Scale must be a float or a tuple of length 2., 
                              instead of length: {len(scale)}"""
            )
        dx, dy = scale
    elif isinstance(scale, float) | isinstance(scale, int):
        dx = dy = scale
    else:
        raise ValueError(
            f"""Scale must be a float or a tuple of length 2, 
                         instead of type {type(scale)}"""
        )
    xmins = np.arange(minx, maxx, dx)
    ymins = np.arange(miny, maxy, dy)
    grid = [
        {"geometry": shapely.box(x, y, x + dx, y + dy), "sub_id": i}
        for i, (x, y) in enumerate(product(xmins, ymins))
    ]

    return gpd.GeoDataFrame(grid)


scale_registry = {}


def register_scale(scale_func: Callable):
    """Register a scale function.

    Register a scale function to the scale_registry. The function should
    take the same arguments as the scale functions and return a float. The
    function should be registered with the '@register_scale' decorator. A scale
    function is called as a metric, but with some additional arguments provided
    (i.e., the variable name and the coefficient function to use). The function
    should return a float.

    Args:
        scale_func (Callable): The scale function to register.
    """
    name = scale_func.__name__

    # Check if the function is already registered
    if name in scale_registry:
        raise ValueError(f"Scale function '{name}' already registered.")

    # Validate the function
    args = list(get_type_hints(scale_func).keys())
    if args != [
        "synthetic_results",
        "synthetic_subs",
        "synthetic_G",
        "real_results",
        "real_subs",
        "real_G",
        "metric_evaluation",
        "var",
        "coef_func",
    ]:
        raise ValueError(
            f"""Scale {scale_func.__name__} requires args 
                         ('synthetic_results', 'synthetic_subs', 'synthetic_G', 
                         'real_results', 'real_subs', 'real_G', 
                         'metric_evaluation', 'var', 'coef_func')."""
        )

    # Add the function to the registry
    scale_registry[name] = scale_func
    return scale_func


@register_scale
def subcatchment(
    synthetic_results: pd.DataFrame,
    synthetic_subs: gpd.GeoDataFrame,
    synthetic_G: nx.Graph,
    real_results: pd.DataFrame,
    real_subs: gpd.GeoDataFrame,
    real_G: nx.Graph,
    metric_evaluation: MetricEvaluation,
    var: str,
    coef_func: Callable,
):
    """Subcatchment scale metric.

    Calculate the coefficient (coef_func) of a variable over time for aggregated
    to real subcatchment scale. The metric produced is the median coef_func
    across all subcatchments.

    Args:
        synthetic_results (pd.DataFrame): The synthetic results.
        synthetic_subs (gpd.GeoDataFrame): The synthetic subcatchments.
        synthetic_G (nx.Graph): The synthetic graph.
        real_results (pd.DataFrame): The real results.
        real_subs (gpd.GeoDataFrame): The real subcatchments.
        real_G (nx.Graph): The real graph.
        metric_evaluation (MetricEvaluation): The metric evaluation parameters.
        var (str): The variable to calculate the coefficient for.
        coef_func (Callable): The coefficient to calculate.

    Returns:
        float: The median coef_func value.
    """
    results = align_by_shape(
        var,
        synthetic_results=synthetic_results,
        real_results=real_results,
        shapes=real_subs,
        synthetic_G=synthetic_G,
        real_G=real_G,
    )

    return median_coef_by_group(results, "sub_id", coef_func=coef_func)


@register_scale
def grid(
    synthetic_results: pd.DataFrame,
    synthetic_subs: gpd.GeoDataFrame,
    synthetic_G: nx.Graph,
    real_results: pd.DataFrame,
    real_subs: gpd.GeoDataFrame,
    real_G: nx.Graph,
    metric_evaluation: MetricEvaluation,
    var: str,
    coef_func: Callable,
):
    """Grid scale metric.

    Classify synthetic nodes to a grid and calculate the coef_func of a variable over
    time for each grid cell. The metric produced is the median coef_func across all
    grid cells.

    Args:
        synthetic_results (pd.DataFrame): The synthetic results.
        synthetic_subs (gpd.GeoDataFrame): The synthetic subcatchments.
        synthetic_G (nx.Graph): The synthetic graph.
        real_results (pd.DataFrame): The real results.
        real_subs (gpd.GeoDataFrame): The real subcatchments.
        real_G (nx.Graph): The real graph.
        metric_evaluation (MetricEvaluation): The metric evaluation parameters.
        var (str): The variable to calculate the coefficient for.
        coef_func (Callable): The coefficient to calculate.

    Returns:
        float: The median coef_func value.
    """
    # Create a grid (GeoDataFrame of polygons)
    scale = metric_evaluation.grid_scale
    grid = create_grid(real_subs.total_bounds, scale)
    grid.crs = real_subs.crs

    # Align results
    results = align_by_shape(
        var,
        synthetic_results=synthetic_results,
        real_results=real_results,
        shapes=grid,
        synthetic_G=synthetic_G,
        real_G=real_G,
    )
    # Calculate coefficient
    return median_coef_by_group(results, "sub_id", coef_func=coef_func)


@register_scale
def outfall(
    synthetic_results: pd.DataFrame,
    synthetic_subs: gpd.GeoDataFrame,
    synthetic_G: nx.Graph,
    real_results: pd.DataFrame,
    real_subs: gpd.GeoDataFrame,
    real_G: nx.Graph,
    metric_evaluation: MetricEvaluation,
    var: str,
    coef_func: Callable,
):
    """Outfall scale metric.

    Calculate the coefficient of a variable for the subgraph that
    drains to the dominant outfall node. The dominant outfall node of the 'real'
    network is calculated by dominant_outfall, while the dominant outfall node of
    the 'synthetic' network is calculated by best_outfall_match.

    Args:
        synthetic_results (pd.DataFrame): The synthetic results.
        synthetic_subs (gpd.GeoDataFrame): The synthetic subcatchments.
        synthetic_G (nx.Graph): The synthetic graph.
        real_results (pd.DataFrame): The real results.
        real_subs (gpd.GeoDataFrame): The real subcatchments.
        real_G (nx.Graph): The real graph.
        metric_evaluation (MetricEvaluation): The metric evaluation parameters.
        var (str): The variable to calculate the coefficient for.
        coef_func (Callable): The coefficient to calculate.

    Returns:
        float: The median coef_func value.
    """
    # Identify synthetic and real arcs that flow into the best outfall node
    sg_syn, syn_outfall = best_outfall_match(synthetic_G, real_subs)
    if len(sg_syn.nodes) == 0:
        # No overlap exists
        return np.inf

    sg_real, real_outfall = dominant_outfall(real_G, real_results)

    allowable_var = ["nmanholes", "diameter", "npipes", "length", "flow", "flooding"]
    if var not in allowable_var:
        raise ValueError(f"Invalid variable {var}. Can be {allowable_var}")

    if var == "nmanholes":
        # Calculate the coefficient based on the number of manholes
        return coef_func(
            np.atleast_1d(sg_real.number_of_nodes()),
            np.atleast_1d(sg_syn.number_of_nodes()),
        )
    if var == "npipes":
        # Calculate the coefficient based on the number of pipes
        return coef_func(
            np.atleast_1d(sg_real.number_of_edges()),
            np.atleast_1d(sg_syn.number_of_edges()),
        )
    if var == "length":
        # Calculate the coefficient based on the total length of the pipes
        return coef_func(
            np.array(sum(nx.get_edge_attributes(sg_real, var).values())),
            np.array(sum(nx.get_edge_attributes(sg_syn, var).values())),
        )
    if var == "diameter":
        # Calculate the coefficient based on the average diameter of the pipes
        return coef_func(
            np.array(list(nx.get_edge_attributes(sg_real, var).values())),
            np.array(list(nx.get_edge_attributes(sg_syn, var).values())),
        )
    if var == "flow":
        # Identify synthetic and real arcs that flow into the best outfall node
        syn_arc = [
            d["id"] for u, v, d in synthetic_G.edges(data=True) if v == syn_outfall
        ]
        real_arc = [d["id"] for u, v, d in real_G.edges(data=True) if v == real_outfall]
    elif var == "flooding":
        # Use all nodes in the subgraphs
        syn_arc = list(sg_syn.nodes)
        real_arc = list(sg_real.nodes)

    # Calculate the coefficient
    df = align_by_id(synthetic_results, real_results, var, syn_arc, real_arc)
    return coef_func(df.value_real, df.value_syn)


def metric_factory(name: str):
    """Create a metric function.

    A factory function to create a metric function based on the name. The first
    part of the name is the scale, the second part is the metric, and the third
    part is the variable. For example, 'grid_nse_flooding' is a metric function
    that calculates the NSE of flooding at the grid scale.

    Args:
        name (str): The name of the metric.

    Returns:
        Callable: The metric function.
    """
    # Split the name
    parts = name.split("_")
    if len(parts) != 3:
        raise ValueError("Invalid metric name. Expected 'scale_metric_variable'")
    scale, metric, variable = parts

    # Get coefficient
    coef_func = coef_registry[metric]

    # Get scale
    func = scale_registry[scale]

    # Validate the metric
    for restriction in restriction_registry.values():
        restriction(scale, metric, variable)

    # Create the metric function
    def new_metric(**kwargs):
        return func(coef_func=coef_func, var=variable, **kwargs)

    new_metric.__name__ = name
    return new_metric


metrics.register(metric_factory("outfall_nse_flow"))
metrics.register(metric_factory("outfall_kge_flow"))
metrics.register(metric_factory("outfall_relerror_flow"))

metrics.register(metric_factory("outfall_relerror_length"))
metrics.register(metric_factory("outfall_relerror_npipes"))
metrics.register(metric_factory("outfall_relerror_nmanholes"))
metrics.register(metric_factory("outfall_relerror_diameter"))

metrics.register(metric_factory("outfall_nse_flooding"))
metrics.register(metric_factory("outfall_kge_flooding"))
metrics.register(metric_factory("outfall_relerror_flooding"))

metrics.register(metric_factory("grid_nse_flooding"))
metrics.register(metric_factory("grid_kge_flooding"))
metrics.register(metric_factory("grid_relerror_flooding"))

metrics.register(metric_factory("subcatchment_nse_flooding"))
metrics.register(metric_factory("subcatchment_kge_flooding"))
metrics.register(metric_factory("subcatchment_relerror_flooding"))


@metrics.register
def nc_deltacon0(synthetic_G: nx.Graph, real_G: nx.Graph, **kwargs) -> float:
    """Run the evaluated metric."""
    return nc_compare(synthetic_G, real_G, "deltacon0", eps=1e-10)


@metrics.register
def nc_laplacian_dist(synthetic_G: nx.Graph, real_G: nx.Graph, **kwargs) -> float:
    """Run the evaluated metric."""
    return nc_compare(synthetic_G, real_G, "lambda_dist", k=None, kind="laplacian")


@metrics.register
def nc_laplacian_norm_dist(synthetic_G: nx.Graph, real_G: nx.Graph, **kwargs) -> float:
    """Run the evaluated metric."""
    return nc_compare(
        synthetic_G.to_undirected(),
        real_G.to_undirected(),
        "lambda_dist",
        k=None,
        kind="laplacian_norm",
    )


@metrics.register
def nc_adjacency_dist(synthetic_G: nx.Graph, real_G: nx.Graph, **kwargs) -> float:
    """Run the evaluated metric."""
    return nc_compare(
        synthetic_G.to_undirected(),
        real_G.to_undirected(),
        "lambda_dist",
        k=None,
        kind="adjacency",
    )


@metrics.register
def nc_vertex_edge_distance(synthetic_G: nx.Graph, real_G: nx.Graph, **kwargs) -> float:
    """Run the evaluated metric.

    Do '1 -' because this metric is similarity not distance.
    """
    return 1 - nc_compare(synthetic_G, real_G, "vertex_edge_distance")


@metrics.register
def nc_resistance_distance(synthetic_G: nx.Graph, real_G: nx.Graph, **kwargs) -> float:
    """Run the evaluated metric."""
    return nc_compare(
        synthetic_G,
        real_G,
        "resistance_distance",
        check_connected=False,
        renormalized=True,
    )


@metrics.register
def bias_flood_depth(
    synthetic_results: pd.DataFrame,
    real_results: pd.DataFrame,
    synthetic_subs: gpd.GeoDataFrame,
    real_subs: gpd.GeoDataFrame,
    **kwargs,
) -> float:
    """Run the evaluated metric."""

    def _f(x):
        return np.trapz(x.value, x.duration)

    syn_flooding = extract_var(synthetic_results, "flooding").groupby("id").apply(_f)
    syn_area = synthetic_subs.impervious_area.sum()
    syn_tot = syn_flooding.sum() / syn_area

    real_flooding = extract_var(real_results, "flooding").groupby("id").apply(_f)
    real_area = real_subs.impervious_area.sum()
    real_tot = real_flooding.sum() / real_area

    return (syn_tot - real_tot) / real_tot


@metrics.register
def kstest_edge_betweenness(synthetic_G: nx.Graph, real_G: nx.Graph, **kwargs) -> float:
    """Run the evaluated metric."""
    syn_betweenness = edge_betweenness_centrality(synthetic_G, weight=None)
    real_betweenness = edge_betweenness_centrality(real_G, weight=None)

    # TODO does it make more sense to use statistic or pvalue?
    return stats.ks_2samp(
        list(syn_betweenness.values()), list(real_betweenness.values())
    ).statistic


@metrics.register
def kstest_betweenness(synthetic_G: nx.Graph, real_G: nx.Graph, **kwargs) -> float:
    """Run the evaluated metric."""
    syn_betweenness = nx.betweenness_centrality(synthetic_G, weight=None)
    real_betweenness = nx.betweenness_centrality(real_G, weight=None)

    # TODO does it make more sense to use statistic or pvalue?
    return stats.ks_2samp(
        list(syn_betweenness.values()), list(real_betweenness.values())
    ).statistic


@metrics.register
def outfall_kstest_diameters(
    real_G: nx.Graph,
    synthetic_G: nx.Graph,
    real_results: pd.DataFrame,
    real_subs: gpd.GeoDataFrame,
    **kwargs,
) -> float:
    """Outfall KStest diameters.

    Calculate the Kolmogorov-Smirnov statistic of the diameters in the subgraph
    that drains to the dominant outfall node. The dominant outfall node of the
    'real' network is calculated by dominant_outfall, while the dominant outfall
    node of the 'synthetic' network is calculated by best_outfall_match.
    """
    # Identify synthetic and real outfall arcs
    sg_syn, _ = best_outfall_match(synthetic_G, real_subs)

    if len(sg_syn.nodes) == 0:
        # No overlap exists
        return np.inf

    sg_real, _ = dominant_outfall(real_G, real_results)

    # Extract the diameters
    syn_diameters = nx.get_edge_attributes(sg_syn, "diameter")
    real_diameters = nx.get_edge_attributes(sg_real, "diameter")
    return stats.ks_2samp(
        list(syn_diameters.values()), list(real_diameters.values())
    ).statistic



"""Preprocessing module for SWMManywhere.

A module to call downloads, preprocess these downloads into formats suitable
for graphfcns, and some other utilities (such as creating a project folder
structure or create the starting graph from rivers/streets).
"""

from __future__ import annotations

import tempfile
from pathlib import Path

import geopandas as gpd
import networkx as nx
import pandas as pd

from swmmanywhere import geospatial_utilities as go
from swmmanywhere import graph_utilities as gu
from swmmanywhere import prepare_data
from swmmanywhere.filepaths import FilePaths
from swmmanywhere.logging import logger


def write_df(df: pd.DataFrame | gpd.GeoDataFrame, fid: Path):
    """Write a DataFrame to a file.

    Write a DataFrame to a file. The file type is determined by the file
    extension.

    Args:
        df (DataFrame): DataFrame to write to a file.
        fid (Path): Path to the file.
    """
    if fid.suffix in (".geoparquet", ".parquet"):
        df.to_parquet(fid)
    elif fid.suffix in (".geojson", ".json"):
        if isinstance(df, gpd.GeoDataFrame):
            df.to_file(fid, driver="GeoJSON")
        else:
            df.to_json(fid)


def prepare_precipitation(
    bbox: tuple[float, float, float, float],
    addresses: FilePaths,
    api_keys: dict[str, str],
    target_crs: str,
    source_crs: str = "EPSG:4326",
):
    """Download and reproject precipitation data."""
    if addresses.bbox_paths.precipitation.exists():
        return
    logger.info(f"downloading precipitation to {addresses.bbox_paths.precipitation}")
    precip = prepare_data.download_precipitation(
        bbox, api_keys["cds_username"], api_keys["cds_api_key"]
    )
    precip = precip.reset_index()
    precip = go.reproject_df(precip, source_crs, target_crs)
    write_df(precip, addresses.bbox_paths.precipitation)


def prepare_elevation(
    bbox: tuple[float, float, float, float], addresses: FilePaths, target_crs: str
):
    """Download and reproject elevation data."""
    if addresses.bbox_paths.elevation.exists():
        return
    logger.info(f"downloading elevation to {addresses.bbox_paths.elevation}")
    with tempfile.TemporaryDirectory(dir=".") as temp_dir:
        fid = Path(temp_dir) / "elevation.tif"
        prepare_data.download_elevation(
            fid,
            bbox,
        )
        go.reproject_raster(target_crs, fid, addresses.bbox_paths.elevation)


def prepare_building(
    bbox: tuple[float, float, float, float], addresses: FilePaths, target_crs: str
):
    """Download and reproject building data."""
    if addresses.bbox_paths.building.exists():
        return

    logger.info(f"downloading buildings to {addresses.bbox_paths.building}")
    prepare_data.download_buildings_bbox(addresses.bbox_paths.building, bbox)

    buildings = gpd.read_parquet(addresses.bbox_paths.building)
    buildings = buildings.to_crs(target_crs)
    write_df(buildings, addresses.bbox_paths.building)


def prepare_street(
    bbox: tuple[float, float, float, float],
    addresses: FilePaths,
    target_crs: str,
    source_crs: str = "EPSG:4326",
    network_types=["drive"],
):
    """Download and reproject street graph.

    Download the street graph within the bbox and reproject it to the UTM zone.
    The street graph is downloaded for all network types in network_types. The
    street graph is saved to the addresses.bbox_paths.street directory.

    Args:
        bbox (tuple[float, float, float, float]): Bounding box coordinates in
            the format (minx, miny, maxx, maxy) in EPSG:4326.
        addresses (FilePaths): Class containing the addresses of the directories.
        target_crs (str): Target CRS to reproject the graph to.
        source_crs (str): Source CRS of the graph.
        network_types (list): List of network types to download. For duplicate
            edges, nx.compose_all selects the attributes in priority of last to
            first. In likelihood, you want to ensure that the last network in
            the list is `drive`, so as to retain information about `lanes`,
            which is needed to calculate impervious area.
    """
    if addresses.bbox_paths.street.exists():
        return
    logger.info(f"downloading network to {addresses.bbox_paths.street}")
    if "drive" in network_types and network_types[-1] != "drive":
        logger.warning(
            """The last network type should be `drive` to retain 
                        `lanes` attribute, needed to calculate impervious area.
                        Moving it to the last position."""
        )
        network_types.pop("drive")
        network_types.append("drive")
    networks = []
    for network_type in network_types:
        network = prepare_data.download_street(bbox, network_type=network_type)
        nx.set_edge_attributes(network, network_type, "network_type")
        networks.append(network)
    street_network = nx.compose_all(networks)

    # Reproject graph
    street_network = go.reproject_graph(street_network, source_crs, target_crs)

    gu.save_graph(street_network, addresses.bbox_paths.street)


def prepare_river(
    bbox: tuple[float, float, float, float],
    addresses: FilePaths,
    target_crs: str,
    source_crs: str = "EPSG:4326",
):
    """Download and reproject river graph."""
    if addresses.bbox_paths.river.exists():
        return
    logger.info(f"downloading river network to {addresses.bbox_paths.river}")
    river_network = prepare_data.download_river(bbox)
    river_network = go.reproject_graph(river_network, source_crs, target_crs)
    gu.save_graph(river_network, addresses.bbox_paths.river)


def run_downloads(
    bbox: tuple[float, float, float, float],
    addresses: FilePaths,
    network_types=["drive"],
):
    """Run the data downloads.

    Run the precipitation, elevation, building, street and river network
    downloads. If the data already exists, do not download it again. Reprojects
    data to the UTM zone.

    Args:
        bbox (tuple[float, float, float, float]): Bounding box coordinates in
            the format (minx, miny, maxx, maxy) in EPSG:4326.
        addresses (FilePaths): Class containing the addresses of the directories.
        network_types (list): List of network types to download.
    """
    target_crs = go.get_utm_epsg(bbox[0], bbox[1])

    # Download precipitation data
    # Currently commented because it doesn't work
    # prepare_precipitation(bbox, addresses, api_keys, target_crs)

    # Download elevation data
    prepare_elevation(bbox, addresses, target_crs)

    # Download building data
    prepare_building(bbox, addresses, target_crs)

    # Download street network data
    prepare_street(bbox, addresses, target_crs, network_types=network_types)

    # Download river network data
    prepare_river(bbox, addresses, target_crs)


def create_starting_graph(addresses: FilePaths):
    """Create the starting graph.

    Create the starting graph by combining the street and river networks.

    Args:
        addresses (FilePaths): Class containing the addresses of the directories.

    Returns:
        nx.Graph: Combined street and river network.
    """
    river = gu.load_graph(addresses.bbox_paths.river)
    nx.set_edge_attributes(river, "river", "edge_type")
    street = gu.load_graph(addresses.bbox_paths.street)
    nx.set_edge_attributes(street, "street", "edge_type")
    return nx.compose(river, street)



"""Test the parameters module."""

from __future__ import annotations

import tempfile
from pathlib import Path

from swmmanywhere.filepaths import FilePaths, filepaths_from_yaml


def test_getattr():
    """Test the __getattr__ method."""
    with tempfile.TemporaryDirectory(dir=".") as temp_dir:
        filepath = Path(temp_dir)
        addresses = FilePaths(
            base_dir=filepath, bbox_bounds=[0, 1, 0, 1], project_name="test"
        )
        assert addresses.model_paths.model_number == 1
        assert addresses.project_paths.base_dir == filepath
        assert addresses.project_paths.project == filepath / "test"
        assert (
            addresses.project_paths.national
            == addresses.project_paths.project / "national"
        )
        assert addresses.bbox_paths.bbox == addresses.project_paths.project / "bbox_1"
        assert addresses.bbox_paths.download == addresses.bbox_paths.bbox / "download"
        assert (
            addresses.bbox_paths.elevation
            == addresses.bbox_paths.download / "elevation.tif"
        )
        assert (
            addresses.bbox_paths.building
            == addresses.bbox_paths.download / "building.geoparquet"
        )
        assert addresses.model_paths.model == addresses.bbox_paths.bbox / "model_1"
        assert (
            addresses.model_paths.subcatchments
            == addresses.model_paths.model / "subcatchments.geoparquet"
        )
        assert (
            addresses.bbox_paths.precipitation
            == addresses.bbox_paths.download / "precipitation.parquet"
        )

        assert addresses.model_paths.model.exists()
        assert addresses.bbox_paths.download.exists()
        assert addresses.project_paths.national.exists()

        addresses.set_bbox_number(2)
        assert addresses.bbox_paths.bbox == addresses.project_paths.project / "bbox_2"
        assert addresses.bbox_paths.download == addresses.bbox_paths.bbox / "download"
        assert addresses.model_paths.model == addresses.bbox_paths.bbox / "model_1"


def test_to_yaml_normal():
    """Test the to_yaml and from_yaml methods."""
    with tempfile.TemporaryDirectory(dir=".") as temp_dir:
        base_dir = Path(temp_dir)
        addresses = FilePaths(base_dir, "test", [0, 1, 0, 1])

        # Write and read
        addresses.to_yaml(base_dir / "test.yaml")
        addresses_ = filepaths_from_yaml(base_dir / "test.yaml")

        # Check
        paths_ = {
            "project_paths": ["base_dir", "project", "national", "national_building"],
            "bbox_paths": [
                "bbox",
                "download",
                "building",
                "precipitation",
                "elevation",
                "river",
                "street",
            ],
            "model_paths": [
                "model",
                "subcatchments",
                "streetcover",
                "edges",
                "nodes",
                "graph",
                "inp",
            ],
        }
        for cat, keys in paths_.items():
            for key in keys:
                assert getattr(getattr(addresses, cat), key) == getattr(
                    getattr(addresses_, cat), key
                )


def test_to_yaml_normal_with_overrides():
    """Test the to_yaml and from_yaml methods."""
    with tempfile.TemporaryDirectory(dir=".") as temp_dir:
        base_dir = Path(temp_dir)
        dummy_file = base_dir / "dummy.txt"
        dummy_file.touch()
        addresses = FilePaths(
            base_dir,
            "test",
            [0, 1, 0, 1],
            elevation=dummy_file,
            inp=dummy_file,
            new_file=dummy_file,
        )

        # Model number override
        addresses.model_paths.model_number = 2

        # Write and read
        addresses.to_yaml(base_dir / "test.yaml")
        addresses_ = filepaths_from_yaml(base_dir / "test.yaml")

        # Check
        paths_ = {
            "project_paths": ["base_dir", "project", "national", "national_building"],
            "bbox_paths": [
                "bbox",
                "download",
                "building",
                "precipitation",
                "elevation",
                "river",
                "street",
            ],
            "model_paths": [
                "model",
                "subcatchments",
                "streetcover",
                "edges",
                "nodes",
                "graph",
                "inp",
            ],
        }
        for cat, keys in paths_.items():
            for key in keys:
                assert getattr(getattr(addresses, cat), key) == getattr(
                    getattr(addresses_, cat), key
                )
        assert addresses_.get_path("new_file") == dummy_file



# -*- coding: utf-8 -*-
"""Created 2024-03-07.

@author: Barnaby Dobson
"""

from __future__ import annotations

from time import time

import pandas as pd

from swmmanywhere import metric_utilities as mu
from swmmanywhere.prepare_data import download_street

if __name__ == "__main__":
    # Download streets
    Gs = download_street((-0.09547, 51.52196, -0.09232, 51.52422))
    Gm = download_street((-0.09547, 51.52196, -0.08876, 51.52661))
    Gl = download_street((-0.09547, 51.52196, -0.08261, 51.53097))
    Gsep = download_street((-0.08261, 51.53097, -0.06946, 51.53718))

    # Define topological metrics
    metrics = [
        "nc_deltacon0",
        "nc_laplacian_dist",
        "nc_laplacian_norm_dist",
        "nc_adjacency_dist",
        "nc_resistance_distance",
        "nc_vertex_edge_distance",
        "kstest_betweenness",
        "kstest_edge_betweenness",
    ]

    # Calculate metrics
    results = []
    for G1, l1 in zip([Gs, Gm, Gl, Gsep], ["a_small", "b_med", "c_large", "d_sep"]):
        for G2, l2 in zip([Gs, Gm, Gl, Gsep], ["a_small", "b_med", "c_large", "d_sep"]):
            for func in metrics:
                start = time()
                val_ = getattr(mu.metrics, func)(synthetic_G=G1, real_G=G2)
                end = time()
                results.append(
                    {"func": func, "val": val_, "time": end - start, "l1": l1, "l2": l2}
                )
    df = pd.DataFrame(results).sort_values(by=["l1", "l2"])

    """
    # Plot heatmap by func
    f,axs = plt.subplots(4,2,figsize=(10,10))
    for (func,grp),ax in zip(df.groupby('func'),axs.reshape(-1)):
        grp = grp.pivot(index='l1',
                    columns= 'l2', 
                    values='val')
        ax.imshow(grp,cmap='Reds')
        ax.set_xticks(range(4))
        ax.set_yticks(range(4))
        ax.set_xticklabels(grp.columns,rotation=90)
        ax.set_yticklabels(grp.index)
        ax.set_title(func)
    plots = Path(__file__).parent
    f.tight_layout()
    f.savefig(plots / 'heatmap.png')
    """



"""Module for graphfcns that change subcatchments."""

from __future__ import annotations

import tempfile
from itertools import product
from pathlib import Path

import geopandas as gpd
import networkx as nx

from swmmanywhere import geospatial_utilities as go
from swmmanywhere import parameters
from swmmanywhere.filepaths import FilePaths
from swmmanywhere.graph_utilities import BaseGraphFunction, register_graphfcn
from swmmanywhere.logging import logger, verbose


@register_graphfcn
class clip_to_catchments(
    BaseGraphFunction,
    required_node_attributes=["x", "y"],
    required_edge_attributes=["length"],
    adds_node_attributes=["community", "basin"],
):
    """clip_to_catchments class."""

    def __call__(
        self,
        G: nx.Graph,
        addresses: FilePaths,
        subcatchment_derivation: parameters.SubcatchmentDerivation,
        **kwargs,
    ) -> nx.Graph:
        """Clip the graph to the subcatchments.

        Derive the subbasins with `subcatchment_derivation.subbasin_streamorder`.
        If no subbasins exist for that stream order, the value is iterated
        downwards and a warning it flagged.

        If `subcatchment_derivation.subbasin_clip_method` is 'subbasin', then
        links between subbasins are removed. If it is 'community', then links
        between communities in different subbasins may be removed based on the
        following method.

        Run Louvain community detection on the street network to create street
        node communities.

        Communities with less than `subcatchment_derivation.subbasin_membership`
        proportion of nodes in a subbasin have their links to all other nodes
        in that subbasin removed. Nodes not in any subbasin are assigned to a
        subbasin to cover all unassigned nodes.

        Community and basin ids are added to nodes mainly to help with debugging.

        Args:
            G (nx.Graph): A graph
            addresses (FilePaths): A FilePaths parameter object
            subcatchment_derivation (parameters.SubcatchmentDerivation): A
                SubcatchmentDerivation parameter object
            **kwargs: Additional keyword arguments are ignored.

        Returns:
            G (nx.Graph): A graph
        """
        G = G.copy()

        # Derive subbasins
        subbasins = go.derive_subbasins_streamorder(
            addresses.bbox_paths.elevation,
            subcatchment_derivation.subbasin_streamorder,
            x=list(nx.get_node_attributes(G, "x").values()),
            y=list(nx.get_node_attributes(G, "y").values()),
            wbt_zip_path=addresses.project_paths.whiteboxtools_binaries_zip,
        )

        if verbose():
            subbasins.to_file(
                str(addresses.model_paths.nodes).replace("nodes", "subbasins"),
                driver="GeoJSON",
            )

        # Extract street network
        street = G.copy()
        street.remove_edges_from(
            [
                (u, v)
                for u, v, d in street.edges(data=True)
                if d.get("edge_type", "street") != "street"
            ]
        )

        # Create gdf of street points
        street_points = gpd.GeoDataFrame(
            G.nodes,
            columns=["id"],
            geometry=gpd.points_from_xy(
                [G.nodes[u]["x"] for u in G.nodes], [G.nodes[u]["y"] for u in G.nodes]
            ),
            crs=G.graph["crs"],
        ).set_index("id")

        # Classify street points by subbasin
        street_points = gpd.sjoin(
            street_points,
            subbasins.set_index("basin"),
            how="left",
        ).rename(columns={"index_right": "basin"})

        if subcatchment_derivation.subbasin_clip_method == "subbasin":
            edges_to_remove = [
                (u, v)
                for u, v in G.edges()
                if street_points.loc[u, "basin"] != street_points.loc[v, "basin"]
            ]
            G.remove_edges_from(edges_to_remove)
            return G

        # Derive road network clusters
        louv_membership = nx.community.louvain_communities(
            street, weight="length", seed=1
        )

        street_points["community"] = 0
        # Assign louvain membership to street points
        for ix, community in enumerate(louv_membership):
            street_points.loc[list(community), "community"] = ix

        # Introduce a non catchment basin for nan
        street_points["basin"] = street_points["basin"].fillna(-1)
        # TODO possibly it makes sense to just remove these nodes, or at least
        # any communities that are all nan

        nx.set_node_attributes(G, street_points["community"].to_dict(), "community")
        nx.set_node_attributes(G, street_points["basin"].to_dict(), "basin")

        # Calculate most percentage of each subbasin in each community
        community_basin = (
            street_points.groupby("community").basin.value_counts().reset_index()
        )
        community_size = street_points.community.value_counts().reset_index()
        community_basin = community_basin.merge(
            community_size, on="community", how="left", suffixes=("_basin", "_size")
        )

        # Normalize
        community_basin["percentage"] = (
            community_basin["count_basin"] / community_basin["count_size"]
        )

        # Identify community-basin combinations where the percentage is less than
        # the threshold
        community_omit = community_basin.loc[
            community_basin["percentage"] <= subcatchment_derivation.subbasin_membership
        ]

        community_basin = community_basin.set_index("basin")

        # Cut links between communities in community_omit and commuities in those
        # basins
        arcs_to_remove = []
        street_points = street_points.reset_index().set_index("basin")
        for idx, row in community_omit.iterrows():
            community_nodes = louv_membership[int(row["community"])]
            basin_nodes = street_points.loc[[row["basin"]], "id"]
            basin_nodes = set(basin_nodes).difference(community_nodes)

            # Include both directions because operation should work on
            # undirected or directed graph
            arcs_to_remove.extend(
                [(u, v, 0) for u, v in product(community_nodes, basin_nodes)]
                + [(v, u, 0) for u, v in product(community_nodes, basin_nodes)]
            )
        G.remove_edges_from(set(G.edges).intersection(arcs_to_remove))
        if G.is_directed():
            subgraphs = len(list(nx.weakly_connected_components(G)))
        else:
            subgraphs = len(list(nx.connected_components(G)))
        logger.info(f"clip_to_catchments has created {subgraphs} subgraphs.")
        return G


@register_graphfcn
class calculate_contributing_area(
    BaseGraphFunction,
    required_edge_attributes=["id", "geometry"],
    adds_edge_attributes=["contributing_area"],
    adds_node_attributes=["contributing_area"],
):
    """calculate_contributing_area class."""

    def __call__(
        self,
        G: nx.Graph,
        subcatchment_derivation: parameters.SubcatchmentDerivation,
        addresses: FilePaths,
        **kwargs,
    ) -> nx.Graph:
        """Calculate the contributing area for each edge.

        This function calculates the contributing area for each edge. The
        contributing area is the area of the subcatchment that drains to the
        edge. The contributing area is calculated from the elevation data.
        Runoff coefficient (RC) for each contributing area is also calculated,
        the RC is calculated using `addresses.bbox_paths.buildings` and
        `addresses.model_paths.streetcover`.

        Also writes the file 'subcatchments.geojson' to
        addresses.model_paths.subcatchments.

        Args:
            G (nx.Graph): A graph
            subcatchment_derivation (parameters.SubcatchmentDerivation): A
                SubcatchmentDerivation parameter object
            addresses (FilePaths): An FilePaths parameter object
            **kwargs: Additional keyword arguments are ignored.

        Returns:
            G (nx.Graph): A graph
        """
        G = G.copy()

        # Carve
        # TODO I guess we don't need to keep this 'carved' file..
        # maybe could add verbose/debug option to keep it
        with tempfile.TemporaryDirectory(dir=".") as temp_dir:
            temp_fid = Path(temp_dir) / "carved.tif"
            go.burn_shape_in_raster(
                [d["geometry"] for u, v, d in G.edges(data=True)],
                subcatchment_derivation.carve_depth,
                addresses.bbox_paths.elevation,
                temp_fid,
            )

            # Derive
            subs_gdf = go.derive_subcatchments(
                G,
                temp_fid,
                wbt_zip_path=addresses.project_paths.whiteboxtools_binaries_zip,
            )
            if verbose():
                subs_gdf.to_file(addresses.model_paths.subcatchments, driver="GeoJSON")

        # Calculate runoff coefficient (RC)
        if addresses.bbox_paths.building.suffix in (".geoparquet", ".parquet"):
            buildings = gpd.read_parquet(addresses.bbox_paths.building)
        else:
            buildings = gpd.read_file(addresses.bbox_paths.building)
        if addresses.model_paths.streetcover.suffix in (".geoparquet", ".parquet"):
            streetcover = gpd.read_parquet(addresses.model_paths.streetcover)
        else:
            streetcover = gpd.read_file(addresses.model_paths.streetcover)

        subs_rc = go.derive_rc(subs_gdf, buildings, streetcover)

        # Write subs
        # TODO - could just attach subs to nodes where each node has a list of subs
        if addresses.model_paths.subcatchments.suffix in (".geoparquet", ".parquet"):
            subs_rc.to_parquet(addresses.model_paths.subcatchments)
        else:
            subs_rc.to_file(addresses.model_paths.subcatchments, driver="GeoJSON")

        # Assign contributing area
        imperv_lookup = subs_rc.set_index("id").impervious_area.to_dict()

        # Set node attributes
        nx.set_node_attributes(G, 0.0, "contributing_area")
        nx.set_node_attributes(G, imperv_lookup, "contributing_area")

        # Prepare edge attributes
        edge_attributes = {
            edge: G.nodes[edge[0]]["contributing_area"] for edge in G.edges
        }

        # Set edge attributes
        nx.set_edge_attributes(G, edge_attributes, "contributing_area")
        return G
